{<                             Disk Index Creator                                        >}
{<                               by Peter Wayner                                         >}
{<                              Copyright 1991 Peter Wayner                      >}
{<                                                                                                  >}

{This is part of the software. See the copyright notice in IndexGlobals.p}
{for details. This file may not be distributed without this copyright notice}
{or separately from the other files. }
unit FileBreaker;
{This unit is takes care of pulling apart a file and sending the words to the Memory Unit.}


interface
uses
IndexGlobals, MemoryUnit;

function ScanFile (fname: filenamestring; DirID: longint; vref: integer; LocalFileNumber: integer): integer;
{Scans the next file and sends all of it out to the memory unit. Word by Word, Dudes.}
{Returns 0 if completed okay and 1 if trouble.}
{LocalFileNumber is the number assigned to this file.}
procedure InitScanner;
{This must be called before scanning begins.}
implementation

var
NowFile: str255;
{The current File's name.}
NowNumber: integer;
{This is the file number that is assigned and stored.}

NowBlock: HParmBlkPtr;
{This is a MacSpecific thing that might be quite amazing by itself.}

NowFref: integer;
{The current file reference. Not currently set to anything when no file is open. Don't trust.}
{Also Macintosh Specific.}
NowHandle: handle;
{This is the block which is hashed. Again Macintosh specific. It is a chunk of }
{memory of FileBlockSize bytes long.}
HandleBase: longint;
{The dereferenced pointer to the base of the handle.}
NowBlockLength: longint;
{Each block read in has a certain number of bytes. This is that number.}
NowBlockPosition: longint;
{The pointer to the next character to be read.}
NowLastBlock: boolean;
{Set to be true when the last block is being scanned.}
NotEOF: boolean;
{Set to be true when the last byte has been scanned.}
tempword: wordstring;
{For processing.}
{Put way out here to save time initializing it each time.}
space: str255;
LocalMax, LocalMin: integer;
{These are lengths that are locally reset so that there is no}
{need to dereference a pointer...}
LocalDisposeExtra: boolean;
{Another local version of what is stored in TheIndex. Set}
{Locally to save a dereferencing...}
TheCurrentValue: longint;
{Used to store four characters at a time.}
StartTick: longint;
{For timing...}
{These next two functions, Less1P and CharAt were borrowed from a larger }
{set of functions distributed freely by Andy Dent as long as they are accompanied}
{by the following notice:}

{ copyright 1990 A.D. Software, free for all use provided this notice retained }
{Andy Dent                                    A.D. Software phone 09 249 2719 }
{    Mac & VAX programmer           94 Bermuda Dve , Ballajura}
{    a_dent@ ennel.cc.uwa.oz           Western Australia 6066}
{    a_dent@fennel.cc.uwa.oz AU ( international ) }

function Less1P (p: ptr): ptr;
inline
$5397, {SUBQ.L #$1, (A7) }
$2E9F;  {MOVE.P (A7)+, (A7) }

function charAt (p: ptr): char;
type
charsInAWord = packed array[1..2] of char;
charsInAWordPtr = ^charsInAWord;
var
w: charsInAWord;
begin
if Odd(longint(p)) then
begin
w := charsInAWordPtr(Less1P(p))^;  { uses Less1P }
charAt := w[2];               { defined in this unit }
end                                   {to get byte before P }
else
charAt := char(p^);
end;  { charAt }


function ScanFile (fname: filenamestring; DirID: longint; vref: integer; LocalFileNumber: integer): integer;
{Scans the next file and sends all of it out to the memory unit.}
{Returns 0 if completed okay and 1 if trouble.}
{LocalFileNumber is the number assigned to this file.}
var
result: integer;
{This handles problems with the text.}


function PullNextBlock: boolean;
{Pulls in the next block of the file into the memory block at NowHandle.}
{ Resets NowBlockPosition and NowBlockLength to be zero and the number of bytes in }
{the block respectively.}
{Returns true if successful and false otherwise.}

{This is also so Macintosh specific that someone porting it would want to rewrite this.}
var
errorcode: OSErr;
{For problems}

begin

errorcode := noErr;
NowBlockLength := FileBlockSize;
errorcode := FSRead(NowFref, NowBlockLength, NowHandle^);
{with NowBlock^ do begin ioPosMode := fsAtMark; ioPosOffset := NowMark; ioReqCount := FileBlocksize; errorcode := PBRead(NowBlock, true);}
{Note that NowBlockLength is set to the number of bytes in the block. This}
{is FileBlockSize at the largest and some number less than it if there aren't that many bytes}
{in the file.}
NowBlockPosition := 0;
if errorcode = noErr then
PullNextBlock := true
else if errorcode = EofErr then
begin
PullNextBlock := true;
NowLastBlock := true
end
else
begin
Error(concat('Macintosh Toolbox File Error #', Num2Str(errorcode)));
PullNextBlock := false
end;

end;{PullNextBlock}

function TestFileForCoolness (fname: filenamestring; DirID: longint): boolean;
{This makes sure that the file is desirable. I.E. Not an executable.}
var
errorcode: OSErr;
{Because computers make mistakes too...}
begin
with NowBlock^ do
begin
iopermssn := fsRdPerm; {Just ask for read permission. Makes life easier.}
ioNamePtr := @fname;
ioDirID := DirId;
ioVrefNum := vref;
ioMisc := nil; {This shouldn't matter.}
end;
errorcode := PBHGetFInfo(NowBlock, false);
if errorcode <> NoErr then
begin
Error(concat('Macintosh File Error #', Num2Str(errorcode)));
TestFileForCoolness := false;
exit(TestFileForCoolness)
end;
if NowBlock^.ioFlFndrInfo.fdType = 'APPL' then
begin
TestFileForCoolness := false;

end
else
TestFileForCoolness := true;
{This is Macintosh specific language for is a file an application, and thus uninteresing to the program.}
end;{TestFileForCoolness}

function OpenFileForScanning (fname: filenamestring; DirID: longint): boolean;
{This function opens a file and loads the first FileBlockSize bytes into}
{memory. This contains so many  MacIntosh Specific operations that }
{anyone using on the system would want to rewrite this entire function}
{and replace it with something that is functionally equivalent.}
var
errorcode, result: integer;
{Used to receive errorcodes.}

begin
NowLastBlock := false;
{errorcode := FSOpen(fName, DirID, NowFref);}
with NowBlock^ do
begin
iopermssn := fsRdPerm; {Just ask for read permission. Makes life easier.}
ioNamePtr := @fname;
ioDirID := DirId;
ioVrefNum := vref;
ioMisc := nil; {This shouldn't matter.}
end;

errorcode := PBHOpen(NowBlock, false);
NowFref := NowBlock^.ioRefNum;
if errorcode <> NoErr then
begin
if errorcode = OpWrErr then
result := FSClose(NowFref);
{    errorcode := FSOpen(fName, DirID, NowFref);}
errorcode := PBHOpen(NowBlock, false);
if errorcode <> NoErr then
begin
Error(concat('Macintosh Toolbox File Error #', Num2Str(errorcode)));
OpenFileForScanning := false
end
else
OpenFileForScanning := PullNextBlock;

end
else
begin
OpenFileForScanning := PullNextBlock
end;

{Try to read in the first block of data and return true if successful.}
end;{OpenFileForScanning}

function NextChar: char;
{This returns the next character to be read in the block. It filters all}
{non-letter data and changes them into spaces. It also changes all lowercase}
{characters to uppercase.The character '*' stands}
{for eof and it is returned at the end. It will represented as EOFChar}
type
byteptr = ^byte;
var
temp: char;
begin
if NowBlockPosition = NowBlockLength then
if NowLastBlock then
begin {End of the line.}
NextChar := ' ';
NotEOF := false;
exit(NextChar);
end
else if not (PullNextBlock) then
begin
NextChar := ' ';
NotEOF := false;
exit(NextChar)
end;
{This should fix everything so there is freshdata.}
temp := charAt(ptr(HandleBase + NowBlockPosition));
{temp := char(BAND($ff, bytePtr(HandleBase + NowBlockPosition)^));}
NowBlockPosition := NowBlockPosition + 1;
if (temp < 'A') | (temp > 'Z') then
if (temp >= 'a') & (temp <= 'z') then
temp := chr(ord(temp) - 32)
else
temp := ' ';
NextChar := temp
end;{NextChar}

function AddNextWordOverLap: integer;
{This function finds the next word and adds it to the file. It}
{does an interesting thing when it finds that the word is longer}
{than MaxTreeDepth, the length of word string. It just keeps deleting}
{a letter from the front of the string and adding another to the end}
{and adding this new word to the file. That is if MaxTreeDepth is 3}
{and it encounters 'Gonzo', then it adds "Gon", then "onz" then "nzo".}
{The results can be anded together.}

{Returns 0 if a-okay, 1 if just an eof and 2 if real trouble.}
var

tempchar: char;

curpos: integer;
{The location being loaded in tempword.}
i: integer;
{For Journeyman's work.}

begin
AddNextWordOverlap := 0;
NotEOF := True;
while NotEOF do
begin
HandleBase := LongInt(NowHandle^) - 3;{This is going to have to be fixed !}
{Pull off all leading spaces...}
tempchar := NextChar;
while (tempchar = ' ') and NotEOF do
tempchar := NextChar;
CurPos := 1;

while (tempchar <> ' ') do
begin
tempword[curpos] := tempchar;
if curpos = LocalMax then
begin
{$IFC TestWordUsage}
WordProfile[LocalMax] := WordProfile[LocalMax] + 1;
{$ENDC}
if not (AddWord(tempword, LocalFileNumber)) then
begin

AddNextWordOverlap := 2; {REAL BIG Trouble like a memory overflow.}
exit(AddNextWordOverlap)
end;
for i := 2 to MaxTreeDepth do
tempword[i - 1] := tempword[i]
end
else
curpos := curpos + 1;
tempchar := NextChar
end;
{$IFC TestWordUsage}
WordProfile[curpos] := WordProfile[curpos] + 1;
{$ENDC}

for i := curpos to MaxTreeDepth do
tempword[i] := ' ';
if curpos > LocalMin then
if not (AddWord(tempword, LocalFileNumber)) then
begin
AddNextWordOverlap := 2; {REAL BIG Trouble like a memory overflow.}
exit(AddNextWordOverlap)
end;
end;

end;{AddNextWordOverlap}

function AddNextWordTruncate: integer;
{This function finds the next word and adds it to the file. It}
{stops when it gets to LocalMax and if the index says, "Dispose"}
{extra, then it does, otherwise it just begins again as if it was}
{the beginning of another word. So "SuperDuper" becomes }
{"Super" and "Duper" if TheIndex^.MaxLettersPerWord is 5.}
{If TheIndex^.DisposeBeyondMax is true, then it only enters }
{"Super" into the database.}

{Returns 0 if a-okay, 1 if just an eof and 2 if real trouble.}

var

tempchar: char;

curpos: integer;
{The location being loaded in tempword.}
i: integer;
{For Journeyman's work.}

begin
AddNextWordTruncate := 0;
tempchar := NextChar;
HandleBase := LongInt(NowHandle^) - 3;{This is going to have to be fixed !}
{Pull off all leading spaces...}
NotEOF := true;
while NotEOF do
begin

while (tempchar = ' ') and NotEOF do
tempchar := NextChar;
CurPos := 1;

while (tempchar <> ' ') & (curpos <= LocalMax) do
begin
tempword[curpos] := tempchar;
curpos := curpos + 1;
tempchar := NextChar
end;

if LocalDisposeExtra then
while (tempchar <> ' ') do
tempchar := NextChar;
{$IFC TestWordUsage}
WordProfile[curpos] := WordProfile[curpos] + 1;
{$ENDC}

for i := curpos to MaxTreeDepth do
tempword[i] := ' ';
if curpos > LocalMin then
if not (AddWord(tempword, LocalFileNumber)) then
begin
AddNextWordTruncate := 2; {REAL BIG Trouble like a memory overflow.}
exit(AddNextWordTruncate)
end;
end

end;{AddNextWordTrunc}



begin
with TheIndex^ do
begin
LocalMax := MaxLettersPerWord;
LocalMin := MinLettersPerWord;
LocalDisposeExtra := DisposeExtra
end;
{$IFC TestWordUsage}
LastWordStart := TheIndex^.TheWords^.UnallocatedStart;
LastListStart := TheIndex^.TheLocations^.UnallocatedStart;
StartTick := tickcount;
{$ENDC}
HLock(NowHandle);
if TestFileForCoolness(fname, DirID) then
begin
if not (OpenFileForScanning(fname, DirId)) then
begin
ScanFile := 1;
exit(ScanFile)
end;

if TheIndex^.OverlapOverKill then
result := AddNextWordOverlap
else
result := AddNextWordTruncate;

if result = 2 then
ScanFile := 1
else
ScanFile := 0;
result := FSClose(NowFref)
end;
HUnLock(NowHandle);
{$IFC TestWordUsage}
writeln('Finished :', fname : 12, ' New Words:', LocalNewWordCount : 6, '/', LocalWordCount : 6);
with TheIndex^ do
writeln('Allocated New Cells W:', (TheWords^.UnAllocatedStart - LastWordStart) div AlphaNodeSize : 5, 'L:', (TheLocations^.UnAllocatedStart - LastListStart) div ListNodeSize : 5);
writeln('Total New Words:', NewWordCount : 6, '/', WordsAdded : 6);
writeln('Percentage L:', (LocalNewWordCount / LocalWordCount) : 6, ' G:', NewWordCount / WordsAdded);
LocalNewWordCount := 0;
LocalWordCount := 0;
writeln('Timing:', tickcount - StartTick);
{$ENDC}
end;{ScanFile}

procedure InitScanner;
{This must be called before scanning begins.}
var
i: integer;
begin
{First initialize a memory location to be used by the scanner.}
if NowBlock = nil then
new(NowBlock);
tempword := '';
space := ' ';
for i := 1 to MaxTreeDepth do
tempword := concat(tempword, space);
if NowHandle = nil then
NowHandle := NewHandle(FileBlockSize);
if NowHandle = nil then
Error('Out of Memory Error.');
end;{InitScanner}

end.