{<                             Disk Index Creator                                        >}
{<                               by Peter Wayner                                         >}
{<                              Copyright 1991 Peter Wayner                      >}
{<                                                                                                  >}

{This is part of the software. See the copyright notice in IndexGlobals.p}
{for details. This file may not be distributed without this copyright notice}
{or separately from the other files. }
unit FileStructure;



{This section is used to maintain a description of the disk in a big heap.}
{The structure will be similar to the alphaheap which is used to maintain a list of }
{letters.}
{That is:}
{Each node will have }
{1) a string with 32 characters that holds either a file name or a directory.}
{2) a pointer to the next file name in alphabetical order.}
{3) if the string represents a directory, then this is a pointer to another list of files.}
{4) Finally, there is a pointer to the predicessor that points to this node. This could either}
{    be a directory pointing down if the current node is the first in a list or it is the file which came before.}

{This file structure must be able to do several things. }
{1) Add a completely new file. This comes as a list of file names using the structure defined in IndexGlobals.}
{2) Delete a file. This should be easy to do.}
{3) Add a directory filled with files by using some MacIntosh Specific code.}
{4) Scan the file list and either find holes where new files must be added to the index}
{     or where ones must be updated.}

interface
uses
IndexGlobals, MemoryUnit, FileBreaker, Deletion;

function InitFileNumberHeap: HeapPtr;
{This sets everything up.}
{The File Number heap is used to store the various file numbers assigned internally.}
function GetScan: boolean;
{Returns Asks for a directory and then passes off the work to the EnumerShell.}
function NewFileNameHeap: FileNameHeapPtr;
{This sets up a new file name heap. It uses the definition in IndexGlobals.}
procedure UpdateIndex;
{This does all the necessary things to update the index. }
procedure LookUpShell;
{Does the basic interface chores of looking up a word in the database.}
procedure PathFromNumber (value: longint; var name: filenamestring; var path: str255);
{This converts a file number from TheNumbers heap into a file in the TheNames heap.}
{The file name is returned in name and the rest is returned in the path string in }
{MacSpecific format. i.e. "Root:SubDirectory One: SubDirectory Two:" with colons in between.}

{$IFC NameTest}
procedure PrintFileNameHeap;
{For testing...}
{$ENDC}

implementation
var


filenamebottom: longint;
{This is a dereferenced pointer to a locked handle representing the bottom of the heap }
{holding the names of the files stored in the index.}


ExtraSF: integer;{ for SetDirectory sample }
{This is zero if my hook is not used.}
{If the hook is activated, then it returns a 1 if  a file name and directoryID are being returned. And this file should be scanned on its own.}
{If 2, then a directory number is being returned and every file in the directory should be scanned.}
{If 3, then the directory number is being returned and it should be scanned RECURSIVELY.}
reply: SFReply;
{ used in all SF samples }
MyCurDir: Longint;
{ for SetDirectory sample }
MyName: string[63];
{The name.}
CurDirStore: LongPtr;
{ pointer to CurDirStore value }
typeList: SFTypeList;
{ typelist for all SF samples }

haveAUX: Boolean;
{ Set to TRUE if A/UX is running }
SFSaveDisk: ShortPtr;
{ pointer to SFSaveDisk value }
flagPtr: ShortPtr;




{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}
{                 First a few utilities...                                                                                  }
{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}

procedure AddToValueList (var listhead: valuelistptr; TheNumber: longint; TheDirId: longint; TheVref: integer);
{Well, the number will almost surely be a reference number to a file}
{and it will almost certainly be added to either ToBeKilled, which }
{would mean all references to it would be removed from the main index}
{or ToBeAdded, which means that this file number will now be added}
{to the pantheon!}
var
NowPointer: valuelistptr;
{Pure Pop for Now People-- ref Nick Lowe.}
LastPointer: valuelistptr;
{Another one that seems to be terminally behind the times...}
tempPointer: valuelistptr;
{For inserting a new one.}
begin
{Remember that the list is sorted.}
if listhead = nil then
begin
new(listhead);
with listhead^ do
begin
value := TheNumber;
DirId := TheDirID;
Vref := TheVref;
next := nil
end;
end
else if TheNumber < listhead^.value then
begin
new(NowPointer);
with NowPointer^ do
begin
value := TheNumber;
DirId := TheDirID;
Vref := TheVref;
next := listhead
end;
listhead := NowPointer;
end
else
begin
LastPointer := listhead;
NowPointer := LastPointer^.next;
while (NowPointer <> nil) & (TheNumber < NowPointer^.value) do
begin
LastPointer := NowPointer;
NowPointer := NowPointer^.next
end;
if NowPointer = nil then
begin
new(NowPointer);
with NowPointer^ do
begin
value := TheNumber;
DirID := TheDirID;
Vref := TheVref;
next := nil
end;
LastPointer^.next := NowPointer
end
else
begin
new(tempPointer);
with temppointer^ do
begin
value := TheNumber;
DirID := TheDirID;
Vref := TheVref;
next := NowPointer;
end;
LastPointer^.next := tempPointer
end
end;

end;{AddToValueList}

procedure DestroyValueList (var listhead: valuelistptr);
{Sends them to the big bit bucket...}
var
NowPointer: valuelistptr;
{Pure Pop for Now People-- ref Nick Lowe.}
begin
while listhead <> nil do
begin
NowPointer := listhead^.next;
dispose(listhead);
listhead := NowPointer
end;
end;{DestroyValueList}


procedure RetireFileNumber (num: longint);
forward;

procedure DestroyAndRetireValueList (var listhead: valuelistptr);
{Sends them to the big bit bucket...}
var
NowPointer: valuelistptr;
{Pure Pop for Now People-- ref Nick Lowe.}
begin
while listhead <> nil do
begin
NowPointer := listhead^.next;
RetireFileNumber(listhead^.value);
dispose(listhead);
listhead := NowPointer
end;
end;{DestroyAndRetireValueList}

procedure ReverseValueList (var listhead: valuelistptr);
{Since we want to add from large numbers to small, this needs to be done occasionally.}
var
temp, last: valuelistptr;
begin
last := nil;
while listhead <> nil do
begin
temp := listhead^.next;
listhead^.next := last;
last := listhead;
listhead := temp;
end;
listhead := last;
end;

{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}
{                 For Creating the initial File Number Structure                                               }
{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}

{--------------------------------------------------------------------}
{ This maintains the file number heap. This is a special construction which was}
{ created because there are only about MaxInt (~32768) file values. It is}
{ quite conceivable that the program will run out of them after awhile. Therefore,}
{ they must be stored in a list so they can be reused.}

{The structure of these things will be easy to understand. It is a list of nodes.}
{There is one  long integer per node. It is a pointer to the next value.}
{When it is assigned, this suddenly becomes a pointer to the value in the }
{file name tree structure defined later.}
{--------------------------------------------------------------------}



function InitFileNumberHeap: HeapPtr;
{This sets everything up.}
var
i: integer;
{For Stepping along.}
value: longint;
{What will be stored.}
location: longptr;
{Where it will be written.}
heapbase: longint;
{An offset quantity.}
tempnumberheap: HeapPtr;
begin
New(tempnumberheap);
with tempnumberheap^ do
begin
Kind := Number;
heapsize := (1 + MaxFileNumber) * FileNodeSize;
heaphndl := NewHandle(heapsize);
if heaphndl = nil then
begin
Error('Not Enough Memory');
InitFileNumberHeap := nil;
exit(InitFileNumberHeap);
end;
UnAllocatedStart := MaxFileNumber;
 {There is one node wasted at the top of the heap.}
HLock(heaphndl);
heapbase := Ord4(heaphndl^);
for i := MinFileNumber + 1 to MaxFileNumber do
begin
location := LongPtr(BSL(i, 2) + heapbase);
location^ := i - 1;
{We're going to go from largest to smallest for speeding up the }
{the insertions into the AddToList...}
end;
HUnLock(heaphndl);
end;
InitFileNumberHeap := tempNumberHeap;
end;{InitFileNumberHeap}

function NewFileNumber (NewFileNameLocation: longint): longint;
{Pops a number off of the FileNumberHeap for new usage.}
{Puts the value of NewFileNameLocation in its place. This will be a pointer}
{to the name in the file name structure. This will allow you }
{to turn a number into a path!.}
var
temp: longint;
location: Longptr;
{temp is the value and location is where it's at.}
begin
with TheIndex^.TheNumbers^ do
begin
HLock(heaphndl);
location := LongPtr(Ord4(heaphndl^) + BSL(UnAllocatedStart, 2));
NewFileNumber := UnAllocatedStart;

UnAllocatedStart := location^;
{We're going to go from largest to smallest for speeding up the }
{the insertions into the AddToList...}
location^ := NewFileNameLocation;
if UnAllocatedStart = MinFileNumber then
begin
Error('Out of FileNumbers');
exit(NewFileNumber)
end;
HUnlock(heaphndl);
end
end;{NewFileNumber}

function RetrieveFileNumberValue (number: longint): longint;
{When a number is allocated, a number is written in its place. This function retrieves it.}
var
location: Longptr;
{temp is the value and location is where it's at.}
begin
with TheIndex^.TheNumbers^ do
begin
HLock(heaphndl);
RetrieveFileNumberValue := LongPtr(Ord4(heaphndl^) + BSL(number, 2))^;
HUnLock(heaphndl)
end;
end;{RetrieveFileNumberValue}

procedure RetireFileNumber (num: longint);
{This sticks another one on the top of the heap.}
{Note: There is NO WAY to make sure that a file hasn't been retired already. So}
{don't just retire numbers with abandon. It will definitely screw up the structure}
{if a file number is retired twice.}
var
NewStart: longint;
{This is what will become the new UnallocatedStart}
location: longptr;
{Where it will go...}
begin
with TheIndex^.TheNumbers^ do
begin
HLock(heaphndl);
location := LongPtr(Ord4(heaphndl^) + BSL(num, 2));{Shift left to multiply by two.}
location^ := UnAllocatedStart;
UnAllocatedStart := num;
HUnlock(heaphndl);
end
end;{RetireFileNumber}



{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}
{                 For Creating the initial File Name Structure                                                  }
{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}

procedure FileNameFill (MemStart: LongInt; BytesToFill: longint; PreviousNode: Longint);
{Fill is used to create a list of unallocated pointers. MemStart is the location }
{that Fill starts with and it keeps going for BytesToFill bytes. PreviousNode is}
{the pointer that is allocated from before. This is included this way}
{because this function will be used both at the beginning and when the heap}
{is increased in size.}
{Note BytesToFill must be a multiple of the node size. Initially this is 8 bytes,}
{but it could be changed.}
var
last: longint;
{This is the last location to be written.}
temp: shortptr;
{This is to put a cap on it.}
twice: longint;
{Just double the FileNameNodeSize.}
begin
twice := 2 * FileNameNodeSize; {Incase it doesn't do constant folding.}
if BytesToFill > 0 then
begin
last := MemStart + BytesToFill;
while MemStart < last do
begin
with FileNameNodePtr(MemStart)^ do
begin
Above := PreviousNode;
next := PreviousNode + twice;
fname := 'No.';
down := EmptyList;
end;
PreviousNode := PreviousNode + FileNameNodeSize;
MemStart := MemStart + FileNameNodeSize;
end;
{Well, we're at the end. Put an EndValue there...}
MemStart := MemStart - FileNameNodeSize;
FileNameNodePtr(MemStart)^.fname := 'End.';

end
end;{FileNameNodeFill}


function NewFileNameHeap: FileNameHeapPtr;
{This sets up a new file name heap. It uses the definition in IndexGlobals.}
label
999; {For errors.}
var
Block: CinfoPBRec;
{A MacSpecific structure for playing with the Disk Operating system.}
RootName: str255;
{The Root of the matter. Or at least the currently mounted disk.}
Err: OSErr;
{A MacSpecific type for sucking up errors from the disk operating system.}

tempheap: FileNameheapptr;
{To hold it until the end.}
begin
New(tempheap);
with tempheap^ do
begin
heapsize := StartingFileNameHeapSize;
heaphndl := NewHandle(StartingFileNameHeapSize);
if heaphndl = nil then
begin
Error('Not Enough Memory');
NewFileNameHeap := nil;
goto 999;
end;
UnallocatedStart := 2 * FileNamenodesize; {There is one node wasted at the top of the heap.}

{Now fill up the memory.}
HLock(heaphndl);
FileNameFill(Ord4(heaphndl^), heapsize, -FileNameNodeSize); {The zero here is because it is the start of the list.}

with Block do
begin
ioNamePtr := @rootName;
ioDrParID := 2;
ioFDirIndex := -1;
ioDrDirId := fsRtDirID;
end;
err := PBGetCAtInfo(@Block, false);


TreeRoot := FileNameNodeSize;

with FileNameNodePtr(Ord4(heaphndl^) + TreeRoot)^ do
begin
fname := rootname;
next := EmptyList;
down := EmptyList;
Above := EmptyList;
Directory := true;
closed := true;
end;
with FileNameNodePtr(Ord4(heaphndl^) + EmptyList)^ do
begin
fname := '';
next := EmptyList;
down := EmptyList;
Above := EmptyList;
Directory := true;
closed := true;
end;


HUnlock(heaphndl);
end;
NewFileNameHeap := tempHeap;
999:
end;{NewFileNameHeap}

function GrowFileNameHeap (FileNameHeap: FileNameheapptr): boolean;
{This baby will expand the size of the heap by the HeapIncrement. It is done because }
{The allocation routine has gotten to the end of the line. It must grow the size of }
{the handle, a macintosh specific operation, and then pad it with more stuff.}
{Returns true if successful and false if unsuccessful.}

{This baby is called when a list building routine reaches the end of the line.}
{Therefore it is not necessary to fix the end of the old heap.}
label
999;{ For trouble}
begin
with FileNameHeap^ do
begin
HUnLock(heaphndl);
SetHandleSize(heaphndl, heapsize + FileNameheapincrement);
if MemError <> NoErr then
begin
Error('Out of Memory!');
GrowFileNameHeap := false;
goto 999;
end;
HLock(heaphndl);
FileNameFill(Ord4(heaphndl^) + heapsize, FileNameheapincrement, heapsize - FileNameNodeSize);

heapsize := heapsize + FileNameheapincrement;
GrowFileNameHeap := true;
end;
999:
end;{GrowFileNameHeap}




{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}
{                 For Finding the way around the structure and Adding Files.                           }
{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}


function RelativeString (st1, st2: str255): integer;
{Returns -1 if st1<st2, 1 if st1>st2 and 0 if =}
{Hopefully faster than the MacSpecific RelString}
var
i: integer;
l1, l2, l: integer;
{The lengths of st1 and st2 respectively. l=min}
begin
l1 := length(st1);
l2 := length(st2);
UprString(st1, false);
UprString(st2, false); {A MacSpecific way of converting them to upppercase.}
if l1 < l2 then
l := l1
else
l := l2;
for i := 1 to l do
if st1[i] <> st2[i] then
if st1[i] < st2[i] then
begin
RelativeString := -1;
exit(RelativeString)
end
else
begin
RelativeString := 1;
exit(RelativeString)
end;
if l1 = l2 then
RelativeString := 0
else if l1 < l2 then
RelativeString := -1
else
RelativeString := 1;
end;{RelativeString}

procedure FileNamePushNode (Node: longint);
{This takes the File Name node at position Node and adds it to the unallocated list.}
var
where: longptr;
{The location being removed.}
pr: longint;
val: integer;
{For Pulling off...}
begin
with TheIndex^.TheNames^ do
begin
where := longptr(Ord4(heaphndl^) + UnAllocatedStart);
FileNameNodePtr(where)^.Above := Node;
where := longptr(Ord4(heaphndl^) + Node);
with FileNameNodePtr(where)^ do
begin
Above := EmptyList;
next := UnAllocatedStart;
fname := 'No.';
end;
UnAllocatedStart := Node;
end
end;{FileNamePushNode}


function FileNamePopNode: longint;
{This baby will remove one node from the unallocated list.}
{This version is local to save locking and unlocking as well as make the error signalling easier.}
{Again this is a special version for dealing with the file name heap.}
var
where: longptr;
{The location being removed.}

begin
with TheIndex^.TheNames^ do
begin
where := longptr(Ord4(heaphndl^) + UnAllocatedStart);
FileNamePopNode := UnAllocatedStart;

with FileNameNodePtr(where)^ do
begin
UnAllocatedStart := next;


{This moves the UnAllocatedStart. If val=EndValue (probably 0, check constant declarations)}
{then ran out of space. Do the grow thing.}
if fname = 'End.' then
if not (GrowFileNameHeap(TheIndex^.TheNames)) then
begin
FileNamePopNode := EmptyList;
end
else
FileNameBottom := Ord4(heaphndl^);
end
end;

end;{FileNamePopNode}


function FindLocation (path: pathlistptr): longint;
{This function looks for a path in the FileName database. If it finds it, then it returns the}
{address of the last part of the name of the string in the heap. The results are left in }
{LastLocation. If it can't find it, it adds it and then}
{returns the address. This is a full service function that borrows plenty of its}
{capabilities from the FindWord. In fact, if I was doing a good job and programming}
{in an object oriented way, then all of this would be structured a bit better. But Object}
{Compilers aren't necessarily as common... Also, the program needs to keep track of a bit}
{more information here because it is going to be sticking nodes in the middle of lists, not just}
{finding them. This should make FindWord a bit more efficient.}



var
OneStepUpLocation: longint;
{This is the level above the current node. It should be empty if it is a root.}
PreviousLocation: longint;
{This is where the one before last location.}
LastLocation: longint;
{The position in the heap after Previous Location.}
NewNode: longint;
{Whenever a new node is inserted, this is where it goes.}
{ Used for insertion of a new node}
{FileNameBottom: longint;}
{The bottom of the locked file name pointer.}
{This is global now.}
temp: longint;
{To save loop dereferencing.}
temprelationship: integer;
{Used to store the results of comparing two strings.}



function LocalFileNamePopNode: longint;
{This baby will remove one node from the unallocated list.}
{This version is local to save locking and unlocking as well as make the error signalling easier.}
{Again this is a special version for dealing with the file name heap.}

{I've made this version local to allow it to EXIT(FindLocation) if an error occurs.}
var
where: longptr;
{The location being removed.}

begin
with TheIndex^.TheNames^ do
begin
where := longptr(Ord4(heaphndl^) + UnAllocatedStart);
LocalFileNamePopNode := UnAllocatedStart;

with FileNameNodePtr(where)^ do
begin
UnAllocatedStart := next;


{This moves the UnAllocatedStart. If val=EndValue (probably 0, check constant declarations)}
{then ran out of space. Do the grow thing.}
if fname = 'End.' then
if not (GrowFileNameHeap(TheIndex^.TheNames)) then
begin
FindLocation := EmptyList;

exit(FindLocation); {The Error has already been signaled by GrowHeap.}
end
end
end;

end;{LocalFileNamePopNode}



function FinishFilePath (path: pathlistptr; StartingNode: longint): longint;
{This baby will add the rest of the PATH, when it can't find it in the heap.  StartingNode is the node of the last}
{node that could be found. The function returns the last node if it succeeeds. Otherwise it }
{returns a value known as EmptyList. For memory}
{lossage problems.}

{Start adding from the first entry in path!}
var
NewNode: longint;
{What will be added.}
begin
NewNode := LocalFileNamePopNode;
temp := FileNameBottom + StartingNode;
FileNameNodePtr(temp)^.down := NewNode;


while (path <> nil) do {There is no maximum depth here!}
with FileNameNodePtr(FileNameBottom + NewNode)^ do
begin
above := StartingNode;
StartingNode := NewNode;
NewNode := LocalFileNamePopNode;
down := NewNode;
next := EmptyList;
Directory := true;
fname := path^.fname;
Closed := true;
{This is always true if it is added here.}
{$IFC NameTest}
writeln('Now Adding in Finishing School:', path^.fname);
{$ENDC}
path := path^.next;
end;
with FileNameNodePtr(FileNameBottom + StartingNode)^ do
begin
Directory := false;
down := 0;

{$IFC NameTest}
writeln('Setting Open:', fname);
{$ENDC}
end;

FileNamePushNode(NewNode);
FinishFilePath := StartingNode

end;{FinishFilePath}



begin

FileNameBottom := Ord4(TheIndex^.TheNames^.heaphndl^);
LastLocation := TheIndex^.TheNames^.TreeRoot;
{For multiple disks sharing the same file, you're going to have to do some work here.}
OneStepUpLocation := EmptyList;
PreviousLocation := EmptyList;

{The Main Loop. }
{The Main loop begins with path set to the next directory/file name in the path. The loop will try to find its location in}
{the tree and if it succeeds it will repeat itself. Otherwise, it will insert the location itself and exit}
{completely.}

while path <> nil do
begin


{$IFC NameTest}
writeln('Looking to Add:', path^.fname);
{$ENDC}
temp := LastLocation + FileNameBottom;

tempRelationship := RelativeString(FileNameNodePtr(FileNameBottom + LastLocation)^.fname, path^.fname);
while (FileNameNodePtr(temp)^.next <> EmptyList) & (tempRelationship = -1) do
begin
{$IFC NameTest}
writeln('Now Going Past:', FileNameNodePtr(FileNameBottom + LastLocation)^.fname);
{$ENDC}
PreviousLocation := LastLocation;
LastLocation := FileNameNodePtr(LastLocation + FileNameBottom)^.next;
temp := LastLocation + FileNameBottom;
tempRelationship := RelativeString(FileNameNodePtr(temp)^.fname, path^.fname);
end;


if tempRelationship = 1 then
begin
NewNode := LocalFileNamePopNode;
with FileNameNodePtr(FileNameBottom + PreviousLocation)^ do
if next = LastLocation then
next := NewNode
else
down := NewNode;
if PreviousLocation = EmptyList then
TheIndex^.TheNames^.TreeRoot := NewNode;
with FileNameNodePtr(FileNameBottom + NewNode)^ do
begin
Above := OneStepUpLocation;
{We Want it Pointing Up To the Previous Level.}
next := LastLocation;
directory := true;
Closed := true;
fname := path^.fname;
{Down will be set by FinishFileName}
{$IFC NameTest}
writeln('Now Adding Invitro:', path^.fname);
{$ENDC}
FindLocation := FinishFilePath(path^.next, NewNode);
Exit(FindLocation)
end
end
else if temprelationship = 0 then
path := path^.next
else if (FileNameNodePtr(temp)^.next = EmptyList) then {Trouble. Note this excludes letter=temp.}
begin
NewNode := LocalFileNamePopNode;
with FileNameNodePtr(FileNameBottom + LastLocation)^ do
begin
next := NewNode;
end;
with FileNameNodePtr(FileNameBottom + NewNode)^ do
begin
Above := OneStepUpLocation;
next := EmptyList;
directory := true;
Closed := true;
fname := path^.fname;
{Down will be set by FinishFileName}
{$IFC NameTest}
writeln('Now Adding At End:', path^.fname, 'Closed:', Closed);
writeln(' Top is : ', FileNameNodePtr(FileNameBottom + OneStepUpLocation)^.fname);

{$ENDC}
FindLocation := FinishFilePath(path^.next, NewNode);

Exit(FindLocation)
end

end;
{Okay, if we reach here, letter should equal w1[i] and we are ready to go down.}
PreviousLocation := LastLocation;
OneStepUpLocation := LastLocation;

if path <> nil then
if (FileNameNodePtr(LastLocation + FileNameBottom)^.down = EmptyList) then {A starting position... Comes at the very beginning.}
begin
{This only occurs at the beginning when it doesn't find the right stuff.}
FindLocation := FinishFilePath(path, LastLocation);
exit(FindLocation);
end
else
LastLocation := FileNameNodePtr(FileNameBottom + LastLocation)^.down


end;

{That was the end of the MainLoop. When, and if, it reaches this point, it has made it through all}
{MaxTreeDepth locations. This means it is either at the end, or it has reached a place}
{where a leaf node should exist, but it doesn't. So Add it.}
FindLocation := LastLocation;

end;{FindLocation}

function PathNameFromDirID (DirID: longint; vRefNum: integer; possiblefilename: filenamestring): pathlistptr;
{see technote 238}
{Possible filename is set to '' if it just wants the path and to the name if it wants to include one.}
var
Block: CinfoPBRec;
directoryName: str255;
fullpathname: pathlistptr;
{This holds the result.}
old: pathlistptr;
{just a temp thing...}
Err: OSErr;
begin


if possiblefilename = '' then
FullPathName := nil
else
begin
new(FullPathName);
FullPathName^.fname := possiblefilename;
FullPathName^.next := nil
end;

with Block do
begin
ioNamePtr := @directoryName;
ioDrParID := DirID;
end;
repeat
with block do
begin
ioVrefNum := vRefNum;
ioFDirIndex := -1;
ioDrDirId := Block.ioDrParID
end;
err := PBGetCAtInfo(@Block, false);
old := FullPathName;
new(FullPathName);
FullPathName^.fname := directoryName;
FullPathName^.next := old;
until (Block.ioDRDirID = fsRtDirID);
PathNameFromDirID := FullPathName;
end;

procedure AddFileName (DirID: longint; vRefNum: integer; possiblefilename: filenamestring);
{This is the core function used to add a new file to the TheIndex^.TheNames. Note }
{That much of the structure (and many of the comments) are similar to }
{AddWord in the MemoryUnit. It is conceivable that I could have developed}
{a good abstraction that would have allowed object oriented programming}
{to handle this better. Unfortunately, those compilers aren't as common. }



var
LastLocation: longint;
{This will hold the LastLetter Examined by FindLocation. This is used for information passing.}
path: pathlistptr;

begin
path := PathNameFromDirID(DirID, vrefnum, possiblefilename);
HLock(TheIndex^.TheNames^.heaphndl);
FileNameBottom := Ord4(TheIndex^.TheNames^.heaphndl^);
LastLocation := FindLocation(path);
AddToValueList(ToBeAdded, NewFileNumber(LastLocation), MyCurDir, vrefnum);
with FileNameNodePtr(FileNameBottom + LastLocation)^ do
begin
Directory := false;
down := NewFileNumber(LastLocation);
end;
HUnLock(TheIndex^.TheNames^.heaphndl);

end;{AddFileName}
{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}
{                 Converting from a File Number to a name and path..                                      }
{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}

procedure PathFromNumber (value: longint; var name: filenamestring; var path: str255);
{This converts a file number from TheNumbers heap into a file in the TheNames heap.}
{The file name is returned in name and the rest is returned in the path string in }
{MacSpecific format. i.e. "Root:SubDirectory One: SubDirectory Two:" with colons in between.}
var
top: longint;
{Used to scan the list.}
begin
with TheIndex^.TheNames^ do
begin
HLock(heaphndl);
FileNameBottom := Ord4(heaphndl^);
with FileNameNodePtr(FileNameBottom + RetrieveFileNumberValue(value))^ do
begin
name := fname;
top := Above;
end;
path := '';


while top <> EmptyList do
with FileNameNodePtr(FileNameBottom + top)^ do
begin
top := Above;
path := concat(fname, ':', path);
end;
HUnLock(heaphndl)
end
end;{PathFromNumber}

procedure WordToFileList (w1: wordstring);
{This does what we've been waiting for. A function that converts a word }
{into a list of files which holds it.}
var
ListStart: longint;
{This is the location within TheNumbers data base that starts the list.}
value: longint;
{The value stored in the list, which is a number signifying a file.}
name: filenamestring;
{What will be looked up.}
path: str255;
{The location of the file in the heirarchy.}
ListBottom: longint;
{The location of the bottom of the list structure.}
len: integer;
begin
writeln('"', w1, '" can be found in these files:');
with TheIndex^ do
begin
len := length(w1);
if len > MaxLettersPerWord then
w1[MaxLettersPerWord + 1] := ' '
else
w1[len + 1] := ' ';
ListStart := FindWord(w1);
HLock(TheLocations^.heaphndl);
ListBottom := Ord4(TheLocations^.heaphndl^);
while ListStart <> Emptylist do
begin
ListMemoryToNode(Longptr(liststart + listbottom), liststart, value);
PathFromNumber(value, name, path);
Writeln(name : 25, ' @', path);
end;
HUnLock(TheLocations^.heaphndl);
end
end;{WordToFileList}

procedure LookUpShell;
{Does the basic interface chores of looking up a word in the database.}
var
wor: str255;
begin
write('Look up: ?');
readln(wor);
UprString(wor, false);
{A MacSpecific function which converts any lower-case letters in the string }
{to uppercase ones.}
WordToFileList(wor);

end;{LookUpShell}
{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}
{                 Automagic Scanning............................................                                         }
{-------------------------------------------------------------------------------}
{-------------------------------------------------------------------------------}

procedure DestroyDirectory (location: longint);
{Occasionally a section of the file structure will prove so utterly devoid of }
{value that a full-scale, complete elimination is all that can be done to maintain}
{the sanity and the vitality of the rest of the file structure. When this happens,}
{DestroyDirectory is the function to call! Do it today.}
var
nextUp: longint;
{The next to die!}
begin
{The TheIndex^.TheNames is assumed to be locked.}

nextUp := FileNameNodePtr(filenamebottom + location)^.down;
FileNamePushNode(location);
while nextUp <> EmptyList do
with FileNameNodePtr(filenamebottom + nextUp)^ do
begin
location := nextUp;
nextUp := next;
{$IFC NameTest}
writeln('$Getting Rid of:', fname);
{$ENDC}

if Directory then
DestroyDirectory(location)
else
begin
AddToValueList(ToBeKilled, down, 0, 0);{The Dir ID doesn't matter here.}
FileNamePushNode(location);
end
end
end;{DestroyDirectory}





function EnumerShell (DirIDToSearch: Longint; recurse: boolean; vref: integer; Scanning: boolean): boolean;
{From Technote 68}
{if Scanning is true, don't flip the Closure of DirIDToSearch, if adding, then do it!}
var
Name: str255;
myCPB: CInfoPBRec;
err: OSErr;
myWDPB: WDPBRec;
TotalFiles, TotalDirectories: integer;

LastLocation, PreviousLocation: longint;
{For moving about the file structure...}



procedure EnumerateCatalog (PreviousLocation, LastLocation: longint; dirIDToSearch: longint; NowClosed: boolean);
{When EnumerateCatalog commences, Last Location points to the first entry in the FileNameStructure}
{in this directory and PreviousLocation points to the entry of directory it will scan.}

{NowClosed controls the addition of new files to the record. If it is closed, well, then it is.}

var
index: integer;
{$IFC NameTest}
LocalName: filenamestring;
{$ENDC}
returned: integer;
LocalOwner: longint;
{This is set so when Adjust Position does it's work, it sets everything up correctly.}

function AdjustPosition (Name: filenamestring; NowClosed: boolean): boolean;
{Well, given a position in the tree and a new file name to insert, do the correct stuff.}
{Issue commands to do deletion if need be.}

{The plan is to find the location of  Name. If the file at LastLocation is larger than}
{it, it should place the new file before it. If it is less, than it should delete all these files}
{until it either gets to the end of the list or it finds a node equal to fname.}

{LastLocation is left pointing to the node with Name in it. Previous location is left pointing}
{To it's predicessor.}
{NowClosed is enclosed here because the compiler is having trouble with scoping.}
{Returns true if it's found it. }
var
TempRelationship: integer;
{This just stores greater (1), less than (-1) or equal (0) result from calling RelativeString}
before, after: longint;
{These are the pointers to the previous and posterious nodes adjacent to the one about to go away.}
begin
tempRelationship := 2;
AdjustPosition := false;
if LastLocation <> EmptyList then
begin
tempRelationship := RelativeString(FileNameNodePtr(LastLocation + FileNameBottom)^.fname, Name);
while tempRelationship = -1 do
{fname is not begin found in the computer now... Delete it.}
begin
with FileNameNodePtr(LastLocation + FileNameBottom)^ do
begin
Before := above;
after := next;
{$IFC NameTest}
writeln('$Getting Rid of:', fname);
{$ENDC}
if Directory then
DestroyDirectory(LastLocation)
else
begin
FileNamePushNode(LastLocation);
AddToValueList(ToBeKilled, down, 0, 0)
end
end;

if FileNameNodePtr(Before + FileNameBottom)^.down = LastLocation then
with FileNameNodePtr(Before + FileNameBottom)^ do
down := after
else
with FileNameNodePtr(PreviousLocation + FileNameBottom)^ do
next := after;

if after <> EmptyList then
begin

LastLocation := after;
tempRelationship := RelativeString(FileNameNodePtr(LastLocation + FileNameBottom)^.fname, Name);
end
else
tempRelationship := 2;{Gets us out of here...}
end
end;

if (NowClosed = false) and ((LastLocation = EmptyList) or (tempRelationship = 1)) then
begin
After := LastLocation;
LastLocation := FileNamePopNode;
if after <> EmptyList then
with FileNameNodePtr(FileNameBottom + After)^ do
above := LocalOwner;

with FileNameNodePtr(PreviousLocation + FileNameBottom)^ do
if index = 1 then
down := LastLocation
else
next := LastLocation;
with FileNameNodePtr(LastLocation + FileNameBottom)^ do
begin
{$IFC NameTest}
writeln('Now Adding by adjustment:', Name);
writeln(' Top is : ', FileNameNodePtr(FileNameBottom + LocalOwner)^.fname);

{$ENDC}
above := LocalOwner;
fname := Name;
Closed := NowClosed;
down := EmptyList;
next := After;
end;
AdjustPosition := true; {We found it! Add it up!}

end;
if temprelationship = 0 then
AdjustPosition := true; {We found it! Add it up!}

end;{AdjustPosition}

begin
index := 1;
LocalOwner := PreviousLocation;

while (err = noErr) do
begin
Name := '';
myCPB.ioFDirIndex := index;
myCPB.ioDrDirID := dirIDToSearch;
err := PBGetCatInfo(@myCPB, false);
{Well here we are with a name...}
if err = noErr then
begin

if AdjustPosition(Name, NowClosed) then
begin
if BitTst(@myCPB.ioFLAttrib, 3) then
{We've found a directory...}
begin
if recurse then
begin
with FileNameNodePtr(LastLocation + FileNameBottom)^ do
begin
Directory := true;
{$IFC NameTest}
writeln('Going Down:', Name);
localname := name;
{$ENDC}
EnumerateCatalog(LastLocation, down, myCPB.ioDrDirID, Closed);
{$IFC NameTest}
writeln('Coming Back From:', localname);
{$ENDC}
end
end;
err := 0
end
else
{We've found a file.}
begin
with FileNameNodePtr(LastLocation + FileNameBottom)^ do
begin
Directory := false;
if Down = EmptyList then
begin
Down := NewFileNumber(LastLocation);
returned := ScanFile(name, DirIdToSearch, Vref, down);
{AddToValueList(ToBeAdded, Down, DirIdToSearch, vref);}
end
else if MyCPB.ioFLMdDat > TheIndex^.LastUpdateDate then
begin
{$IFC NameTest}
writeln('Need to update:', fname);
{$ENDC}
with FileNameNodePtr(LastLocation + FileNameBottom)^ do
begin
AddToValueList(ToBeKilled, down, 0, 0);
AddToValueList(ToBeAdded, down, DirIdToSearch, vref)
end;
end
end;

end;

PreviousLocation := LastLocation;
LastLocation := FileNameNodePtr(LastLocation + FileNameBottom)^.next
end;
index := index + 1;

end
end
end;

begin

HLock(TheIndex^.TheNames^.heaphndl);
FileNameBottom := Ord4(TheIndex^.TheNames^.heaphndl^);
LastLocation := FindLocation(PathNameFromDirID(DirIDToSearch, vref, ''));
with FileNameNodePtr(FileNameBottom + LastLocation)^ do
begin
Directory := true;
if not (Scanning) then
Closed := false
{Since we are not scanning we want to flip this bit right here so it will now be a full directory!}
end;
{FindLocation should set Closed to true iff it has added a new location.}
{if it has merely located a new location, then it stops.}


myWDPB.ioNamePtr := nil;
myWDPB.ioWDDirID := DirIdToSearch;
myWDPB.ioVrefNum := vref;
err := PBHGetVol(@myWDPB, false);
with MyCPB do
begin
ioNamePtr := @Name;
ioVrefNum := myWDPB.ioWDVrefNum;
end;
if LastLocation <> EmptyList then
begin
{$IFC NameTest}
with FileNameNodePtr(FileNameBottom + LastLocation)^ do
begin
writeln('About to descend from :', fname, 'at ', LastLocation);
end;
{$ENDC}
with FileNameNodePtr(FileNameBottom + LastLocation)^ do
EnumerateCatalog(LastLocation, down, DirIdToSearch, closed);
end
end;

const
rGetDirectoryDLOG = 1002;

firstTime = -1;  {the first time our hook is called, it is passed a -1}

function MyGetDirHook (item: integer; dPtr: DialogPtr): integer;

const
{ Equates for the items that I've added }
getDirButton = 11;
getDirRecursively = 12;

var
messageTitle: str255;
junk: integer;
err: OSErr;
begin
{ By default, return the item passed to us. }
MyGetDirHook := item;

case item of
firstTime: 
begin

{ Read in the prompt string from the resource fork, and initialize }
{ CurDirValid to FALSE. }

{GetIndString(messageTitle, rStrMisc, 3);}
{GetDItem(dPtr, getDirMessage, kind, h, r);}
{SetIText(h, messageTitle);}
ExtraSF := 0;
end;
getDirButton: 
begin
{if LONGINT(reply.fType) <> 0 then begin}
myGetDirHook := getCancel; {Fake it out!}
err := GetVol(@MessageTitle, junk);
{if reply.fname = '' then }
{We've got a directory.}
{begin}
reply.vrefnum := -1 * SFSaveDisk^;
ExtraSF := 2;
MyCurDir := CurDirStore^;
{MyCurDir := LONGINT(reply.fType);}
{Use the stuff in brackets if you want the current directory, not the one selected.}



{end}
{else}
{begin MyCurDir := CurDirStore^; MyName := reply.fname; ExtraSF := 1; myGetDirHook := getOpen;}
 {Fake it out!}

{end;}
{end}
end;
getDirRecursively: 
begin
ExtraSF := 3;
MyCurDir := CurDirStore^;
reply.vrefnum := -1 * SFSaveDisk^;

myGetDirHook := getCancel; {Fake it out!}


end;
otherwise
;
end;
end;





procedure AddEm (var listofValues: valuelistptr);
{When AddEm starts, it puts em all in the system.}

{what needs to be done here: find out the file's name. UNLOCK the filenameheap.Do the add. Relock and begin.}
var
temp: valuelistptr;
returned: integer;
FileNameBottom: longint;
ScanMe: filenamestring;
{So we can unlock the FileName Block}
Block: CinfoPBRec;
Err: OSErr;
junk: str255;
{for kicks.}
begin
temp := listofvalues;
returned := 0;
while (temp <> nil) & (returned < 2) do
with temp^, TheIndex^.TheNames^ do
begin
HLock(heaphndl);
FileNameBottom := Ord4(heaphndl^);
ScanMe := FileNameNodePtr(FileNameBottom + RetrieveFileNumberValue(value))^.fname;
HUnLock(heaphndl);

{Unlock it here because it needs to be done to save memory movement during additon.}

err := 0;
if err = NOErr then
returned := ScanFile(ScanMe, DirId, Vref, value)
else
returned := 2;
temp := temp^.next;
end;
DestroyValueList(listofvalues);
end;{AddEm}

function GetScan: boolean;
{Returns a list of files to be scanned by the big one.}
{Of course, if it returns nil, then forget about it.}
const
rGetDirectoryDLOG = 1002;

firstTime = -1;  {the first time our hook is called, it is passed a -1}
rStrMisc = 256;
kHWCfgFlags = $B22;{ Used to see if A/UX is running}
kSFSaveDisk = $214;{ Negative of current volume refnum [WORD]}
kCurDirStore = $398;{ DirID of current directory [LONG]}

var

temp: str255;
{Just to suck up the stuff coming from the SetVol}
err: OSErr;
{For the same reason as temp.}
where: point;
tempptr: pathlistptr;
begin
InitScanner;
{Get's the memory ready for scanning.}
ExtraSF := 0;
flagPtr := ShortPtr(kHWCfgFlags);
if BAnd(flagPtr^, $0200) <> 0 then
begin
haveAUX := TRUE;
end
else
begin
haveAUX := FALSE;
end;
SFSaveDisk := ShortPtr(kSFSaveDisk);
CurDirStore := LongPtr(kCurDirStore);
{This is zero if my hook is not used.}
{If the hook is activated, then it returns a 1 if  a file name and directoryID are being returned. And this file should be scanned on its own.}
{If 2, then a directory number is being returned and every file in the directory should be scanned.}
{If 3, then the directory number is being returned and it should be scanned RECURSIVELY.}
where.h := 3;
where.v := 10;
SFPGetFile(where, 'Yak', nil, -1, typeList, @MyGetDirHook, reply, rGetDirectoryDLOG, nil);{location}
{vestigial string}
{fileFilter}
{numtypes; -1 means all}
{array to types to show}
{dlgHook}
{record for returned values}


case ExtraSF of
0, 1: 
if reply.good then
begin
with reply do
begin
AddFileName(MyCurDir, vrefnum, fname);


end;
{GetScan := (ScanFile(NowFile, NowVref, NowNumber) = 0);}
{$IFC GraphHeapGrowth}
GraphData;
{$ENDC}
end
else
GetScan := false;
2: 
begin
err := SetVol(@temp, reply.vrefnum);
tempptr := PathNameFromDirID(MyCurDir, reply.vrefnum, '');
GetScan := EnumerShell(MyCurDir, false, reply.vrefnum, false);
{False, because it is not recursively and false because we are ADDING new!}
end;
3: 
begin
err := SetVol(@temp, reply.vrefnum);
GetScan := EnumerShell(MyCurDir, true, reply.vrefnum, false);
{True because it will be done recursively and false because we are ADDING new!}
end
end;
ReverseValueList(ToBeAdded);

AddEm(ToBeAdded);
TheIndex^.FileChanged := true;
end;{GetScan}




function MissingVolume (name: str255): integer;
{Takes care of the error when a volume is missing... 0 means ignore, 1 means try again.}
var
answer: integer;
{For the sake of it.}
begin
writeln(name : 32, ' is not mounted. 0 to ignore, 1 to try again:');
readln(answer);
if answer = 1 then
MissingVolume := 1
else
MissingVolume := 0;
end;

procedure UpdateIndex;
{This does all the necessary things to update the index. }
const
RootDir = 2;
{I think.}
var
tops: longint;
localstring: str255;
vref: integer;
{For settting the volume.}
err: OSErr;
{Because things go wrong. Especially here.}
begin
{What this basically does is scan all the volume names in the name heap, sees if they're mounted}
{and then calls }
{There needs to be something called an "Add new directories bit on a directory. If it is not set, then}
{the scanner skips it, if it IS set, well, then it will add new directories that come along. It is recursively}
{set...}
with TheIndex^.TheNames^ do
begin
HLock(heaphndl);
FileNameBottom := Ord4(heaphndl^);
tops := TreeRoot;
while tops <> EmptyList do
with FileNameNodePtr(FileNameBottom + tops)^ do
begin
vref := 0;
localstring := fname;
err := SetVol(@localstring, vref);
if err <> NoErr then
begin
if MissingVolume(localstring) = 0 then
tops := next
{Try to remount it if they answer 1.}
end
else
begin
if EnumerShell(RootDir, true, vref, true) then
tops := next
else
tops := EmptyList;
end
end;
{Now do the add and delete stuff...}
HUnLock(heaphndl);

TheIndex^.FileChanged := true;

if ToBeKilled <> nil then
begin
DeleteBadFilesFromHeap(TheIndex, ToBeKilled);
DestroyValueList(ToBeKilled)
end;
if ToBeAdded <> nil then
begin
InitScanner;

AddEm(ToBeAdded)
end;
GetDateTime(TheIndex^.LastUpdateDate);
{And set the date to be current.}
end

end;{UpdateIndex}


{$IFC nametest}
procedure PrintFileNameHeap;
{For testing...}
const
indentincrem = 3;
var
FileNameBottom: longint;

procedure CurseAndRecurse (location: longint; indent: integer);
{This does the printing...}
var
i: integer;
begin
while location <> EmptyList do
with FileNameNodePtr(FileNameBottom + location)^ do
begin
for i := 1 to indent do
write(' ');
if Directory then
begin
write('At Dir:', fname : 20);
if closed then
writeln('::Closed')
else
writeln;
CurseAndRecurse(down, indent + indentincrem)
end
else
writeln('At:', fname : 20);
location := next;
end;
end;{CurseAndRecurse}
begin
with TheIndex^.TheNames^ do
begin
HLock(heaphndl);
FileNameBottom := Ord4(heaphndl^);
CurseAndRecurse(TreeRoot, 2);
HUnLock(heaphndl)
end
end;{PrintFileNameHeap}
{$ENDC}
end.