{<                             Disk Index Creator                                        >}
{<                               by Peter Wayner                                         >}
{<                              Copyright 1991 Peter Wayner                      >}
{<                                                                                                  >}

{This is part of the software. See the copyright notice in IndexGlobals.p}
{for details. This file may not be distributed without this copyright notice}
{or separately from the other files. }


{This is the central control of the system.}
{It should take orders from a text order processor and then dispatch them.}
{It should be able to do the following things:}
{1) Create a New Index. }
{2) Add files to the Index.}
{3) Scan the index for changes.}
{4) Process queries. Including and's and ors.}
{5) Save and recall Indices.}
{}


unit InAndOut;


interface

uses
IndexGlobals, Deletion, FileStructure, MemoryUnit;

function CreateNewIndex: IndexPtr;
{Builds up a new Index.}

function WriteOutIndex (ThisIndex: indexptr; filename: str255; vref: integer): boolean;
{This Tries to save an index on the disk.}
function OpenUp: boolean;
{Opens up a Indicus file and reads it in.}

procedure DoStartup;
{This includes all of the text that is essentially copied from Macintosh}
{Revealed. That book on my shelf. }
{***** This is the booger machine for handling different things coming in.}

procedure SaveAsAnother;
{Miss Mascarade sez that we should go forth and multiply. So why save the old data. Make it new again.}

procedure Revert;
{Not much here, just read in the stuff and try again. Oh, yeah, ask the bloke if he really wants to do it.}

procedure NewName;
{Just get a new name for now...}
implementation

const
TheExtra = 256;
{This is the number of bytes that need to be added to the front of the Index Files on disk.}


function CreateNewIndex: IndexPtr;
{This sets everything up nice and straightforward. It does all of the necessary}
{initialization.}
var
tempIndexPtr: IndexPtr;

procedure GetLost;
{Memory Errors}
begin
dispose(tempIndexPtr);
CreateNewIndex := nil;
exit(CreateNewIndex)
end;{GetLost}
begin
new(tempIndexPtr);
with tempIndexPtr^ do
begin
TheNames := NewFileNameHeap;
if TheNames = nil then
GetLost;
TheNumbers := InitFileNumberHeap;
if TheNumbers = nil then
GetLost;
TheWords := NewHeap(alpha);
if TheWords = nil then
GetLost;
TheLocations := NewHeap(list);
if TheLocations = nil then
GetLost;
TheRoots := InitRoots(TheWords);
GetDateTime(LastUpdateDate);
{   GetDateTime returns in the secs parameter the contents of the low-}
{memory location in which the date and time is stored; if the date and time is properly}
{set , secs will contain the number of seconds between midnight , January 1 , 1904 and the time that the }
MaxLettersPerWord := 8;
{A default depth set in the globals.}
MinLettersPerWord := 4;
{I think 2 or 3 might make a better default.}
DisposeExtra := true;
OverLapOverKill := false;
{This is probably the most spacesaving.}
FileChanged := false;
end;
CreateNewIndex := tempIndexPtr;
end;{CreateNewIndex}

function EmptyIndex: indexptr;
{creates a new, empty index for the purpose of reading in stuff to memory.}
var
tempIndex: indexptr;
begin
new(tempIndex);
with tempIndex^ do
begin
new(TheWords);
with TheWords^ do
begin
UnAllocatedStart := 0;
heapsize := 0;
heaphndl := nil
end;
new(TheNumbers);
with TheWords^ do
begin
UnAllocatedStart := 0;
heapsize := 0;
heaphndl := nil
end;
new(TheLocations);
with TheWords^ do
begin
UnAllocatedStart := 0;
heapsize := 0;
heaphndl := nil
end;
new(TheNames);
with TheNames^ do
begin
UnAllocatedStart := 0;
TreeRoot := 0;
heapsize := 0;
heaphndl := nil
end;
TheRoots := InitRoots(TheWords);
GetDateTime(LastUpdateDate);
{   GetDateTime returns in the secs parameter the contents of the low-}
{memory location in which the date and time is stored; if the date and time is properly}
{set , secs will contain the number of seconds between midnight , January 1 , 1904 and the time that the }
MaxLettersPerWord := MaxTreeDepth;
{A default depth set in the globals.}
MinLettersPerWord := 1;
{I think 2 or 3 might make a better default.}
FileChanged := false;
end
end;{EmptyIndex}

procedure BlowOutOldIndex (var TheIndex: IndexPtr);
{Gets rid of everything. Var to really clean up.}
label
17;
var
err: OSErr;
{For when things go wrong.}
begin
with TheIndex^ do
begin
DisposHandle(TheNames^.heaphndl);

if MemError <> noErr then
begin
Error(concat('OS FileErr of Type #', Num2Str(Err)));
goto 17;
end;
DisposHandle(TheNumbers^.heaphndl);
if MemError <> noErr then
begin
Error(concat('OS FileErr of Type #', Num2Str(Err)));
goto 17;
end;
DisposHandle(TheWords^.heaphndl);
if MemError <> noErr then
begin
Error(concat('OS FileErr of Type #', Num2Str(Err)));
goto 17;
end;
DisposHandle(TheLocations^.heaphndl);
if MemError <> noErr then
begin
Error(concat('OS FileErr of Type #', Num2Str(Err)));
goto 17;
end;
end;
dispose(TheIndex);
17:
TheIndex := nil;

end;

function WriteOutIndex (ThisIndex: indexptr; filename: str255; vref: integer): boolean;
{This Tries to save an index on the disk.}
var
err: OSErr;

procedure ProcessError (err: OSErr);
{Does nothing if it is a simple error, otherwise exits cleanly.}
begin
if err <> noErr then
begin
case err of
bdNamErr: 
Error('Bad file name.');
extFSErr: 
Error('External file system error.');
fnfErr: 
Error('File not found.');
ioErr: 
Error(' I / O error.');
nsvErr: 
Error(' No such volume mounted.');
dskFulErr: 
Error(' Disk full.');
fLckdErr: 
Error(' File locked.');
fnOpnErr: 
Error(' File not open.');
rfNumErr: 
Error(' Bad reference number.');
vLckdErr: 
Error(' Software volume locked.');

wrPermErr: 
Error(' Read / write permission does not allow writing . ')

end;
writeOutIndex := false;
exit(WriteOutIndex)
end

end;{ProcessError}

var
ActualBytes, NeededBytes: longint;
{How much is needed by the writing out process...}
fref: integer;
{This is where it is going to go.}
amount: longint;
{Something for adjusting this...}
begin
WriteOutIndex := true;
{Unless something goes wrong.}
with ThisIndex^ do
begin
ProcessError(FSOpen(filename, vref, fref));
ProcessError(GetEOF(fref, ActualBytes));
NeededBytes := TheWords^.heapsize + TheNumbers^.heapsize + TheNames^.heapsize + TheLocations^.heapsize;
NeededBytes := NeededBytes + sizeof(TheRoots) + 11 * sizeof(NeededBytes);
{There should be 11 long integers stored away.}
if NeededBytes > ActualBytes then
begin
ActualBytes := NeededBytes - ActualBytes;
ProcessError(Allocate(fref, ActualBytes))
end;
amount := sizeof(NeededBytes);
ProcessError(FSWrite(fref, amount, @NeededBytes));
amount := sizeof(LastUpdateDate);
ProcessError(FSWrite(fref, amount, @LastUpdateDate));
amount := sizeof(TheRoots);
ProcessError(FSWrite(fref, amount, @TheRoots));

{TheWords}
with TheWords^ do
begin
amount := sizeof(heapsize);
ProcessError(FSWrite(fref, amount, @heapsize));
amount := sizeof(UnAllocatedStart);
ProcessError(FSWrite(fref, amount, @UnAllocatedStart));
HLock(heaphndl);
ProcessError(FSWrite(fref, heapsize, heaphndl^));
HUnlock(heaphndl);
end;

{TheLocations}
with TheLocations^ do
begin
amount := sizeof(heapsize);
ProcessError(FSWrite(fref, amount, @heapsize));
amount := sizeof(UnAllocatedStart);
ProcessError(FSWrite(fref, amount, @UnAllocatedStart));
HLock(heaphndl);
ProcessError(FSWrite(fref, heapsize, heaphndl^));
HUnlock(heaphndl);
end;

{TheNames}
with TheNames^ do
begin
amount := sizeof(heapsize);
ProcessError(FSWrite(fref, amount, @heapsize));
amount := sizeof(TreeRoot);
ProcessError(FSWrite(fref, amount, @TreeRoot));
amount := sizeof(UnAllocatedStart);
ProcessError(FSWrite(fref, amount, @UnAllocatedStart));
HLock(heaphndl);
ProcessError(FSWrite(fref, heapsize, heaphndl^));
HUnlock(heaphndl);
end;

{TheNumbers}
with TheNumbers^ do
begin
amount := sizeof(heapsize);
ProcessError(FSWrite(fref, amount, @heapsize));
amount := sizeof(UnAllocatedStart);
ProcessError(FSWrite(fref, amount, @UnAllocatedStart));
HLock(heaphndl);
ProcessError(FSWrite(fref, heapsize, heaphndl^));
HUnlock(heaphndl);
end;

end;
ProcessError(GetFPos(fref, amount));
ProcessError(SetEof(fref, amount));
ProcessError(FSClose(fref));

end;{WriteOutIndex}

function ReadInIndex (var ThisIndex: IndexPtr; filename: str255; vref: integer): boolean;
{Gets the stuff and tries to get enough memory to hold it...}
procedure ProcessError (err: OSErr);
{Does nothing if it is a simple error, otherwise exits cleanly.}
begin
if err <> noErr then
begin
case err of
bdNamErr: 
Error('Bad file name.');
extFSErr: 
Error('External file system error.');
fnfErr: 
Error('File not found.');
ioErr: 
Error(' I / O error.');
nsvErr: 
Error(' No such volume mounted.');
dskFulErr: 
Error(' Disk full.');
fLckdErr: 
Error(' File locked.');
fnOpnErr: 
Error(' File not open.');
rfNumErr: 
Error(' Bad reference number.');
vLckdErr: 
Error(' Software volume locked.');
wPrErr: 
Error(' Hardware volume lock.');
wrPermErr: 
Error(' Read / write permission does not allow writing . ')

end;
exit(ReadInIndex)
end

end;{ProcessError}
var
NeededBytes: longint;
{This is the amount of space the file will take.}
fref: integer;
{The currently accessed file.}
amount: longint;
{The bytes to be stored. Used plenty.}
begin
ReadInIndex := false;
{Unless something goes wrong.}
if ThisIndex = nil then
ThisIndex := CreateNewIndex;
{Make a new one...}
with ThisIndex^ do
begin
ProcessError(FSOpen(filename, vref, fref));

amount := sizeof(NeededBytes);
ProcessError(FSRead(fref, amount, @NeededBytes));
if NeededBytes > FreeMem then
begin
Error('There is not enough memory to open this index.');
exit(ReadInIndex);
end;
amount := sizeof(LastUpdateDate);
ProcessError(FSRead(fref, amount, @LastUpdateDate));
amount := sizeof(TheRoots);
ProcessError(FSRead(fref, amount, @TheRoots));

with TheWords^ do
begin
amount := sizeof(heapsize);
ProcessError(FSRead(fref, amount, @heapsize));
amount := sizeof(UnAllocatedStart);
ProcessError(FSRead(fref, amount, @UnAllocatedStart));
if heaphndl = nil then
heaphndl := NewHandle(heapsize)
else
SetHandleSize(heaphndl, heapsize);
if heaphndl = nil then
Error('Not Enough Memory.');
HLock(heaphndl);
ProcessError(FSRead(fref, heapsize, heaphndl^));
end;

with TheLocations^ do
begin
amount := sizeof(heapsize);
ProcessError(FSRead(fref, amount, @heapsize));
amount := sizeof(UnAllocatedStart);
ProcessError(FSRead(fref, amount, @UnAllocatedStart));
if heaphndl = nil then
heaphndl := NewHandle(heapsize)
else
SetHandleSize(heaphndl, heapsize);
if heaphndl = nil then
Error('Not Enough Memory.');
HLock(heaphndl);
ProcessError(FSRead(fref, heapsize, heaphndl^));
end;

with TheNames^ do
begin
amount := sizeof(heapsize);
ProcessError(FSRead(fref, amount, @heapsize));
amount := sizeof(TreeRoot);
ProcessError(FSRead(fref, amount, @TreeRoot));
amount := sizeof(UnAllocatedStart);
ProcessError(FSRead(fref, amount, @UnAllocatedStart));
if heaphndl = nil then
heaphndl := NewHandle(heapsize)
else
SetHandleSize(heaphndl, heapsize);
if heaphndl = nil then
Error('Not Enough Memory.');
HLock(heaphndl);
ProcessError(FSRead(fref, heapsize, heaphndl^));
end;

with TheNumbers^ do
begin
amount := sizeof(heapsize);
ProcessError(FSRead(fref, amount, @heapsize));
amount := sizeof(UnAllocatedStart);
ProcessError(FSRead(fref, amount, @UnAllocatedStart));
if heaphndl = nil then
heaphndl := NewHandle(heapsize)
else
SetHandleSize(heaphndl, heapsize);
if heaphndl = nil then
Error('Not Enough Memory.');
HLock(heaphndl);
ProcessError(FSRead(fref, heapsize, heaphndl^));
end;
end;
ProcessError(FSClose(fref));
ReadInIndex := true;

end;{ReadInIndex}

function StandardFile (opCode: StandardType; oldName: Str255; fType: OSType; var vRef: integer): Str255;
var
where: Point;
reply: SFReply;
textType: SFTypeList;
begin
where.h := 3;
where.v := 55;
textType[0] := 'TEXT';
if opCode = StandardGet then
SFGetFile(where, '', nil, -1, textType, nil, reply)
else
SFPutFile(where, '', oldName, nil, reply);
with reply do
if not good then
StandardFile := ''
else
begin
StandardFile := fName;
vRef := vRefNum
end
end;

function OpenUp;{:boolean}
{Opens up a Indicus file and reads it in.}
label
15;
var
fname: str255;
vref: integer;
{The stuff of which files are made...}
result: integer;
{What the dude says...}
begin
OpenUp := false;
if TheIndex^.FileChanged then
begin
result := 3;
while (result > 2) or (result < 0) do
begin
write('Save the File? (0=yes, 1=no, 2=cancel operation)');
readln(result)
end;
if result = 0 then
begin
if TheIndex^.OpenFileName = DefaultFileName then
SaveAsAnother{Well, we should get the name...}
else if not (WriteOutIndex(TheIndex, TheIndex^.OpenFileName, TheIndex^.OpenVref)) then
goto 15;
end
else if result = 2 then
goto 15
end;
fname := StandardFile(StandardGet, '', IndicusFileType, TheIndex^.Openvref);
if fname = '' then
goto 15;

if not (ReadInIndex(TheIndex, fName, TheIndex^.OpenVref)) then
goto 15;

TheIndex^.OpenFileName := fname;


15:
end;{OpenUp}

procedure DoStartup;
{This includes all of the text that is essentially copied from Macintosh}
{Revealed. That book on my shelf. }
{***** This is the booger machine for handling different things coming in.}

var
theMessage: integer;
nDocs: integer;{Number of docs selected in Finder.}
thisDoc: integer;
docInfo: AppFile;{Startup information about one document.}
ignore: integer;

begin
CountAppFiles(theMessage, nDocs);
if theMessage = AppPrint then
begin
Error('The Printing function is not enabled.');
ExitToShell
end
else if nDocs > 0 then
begin
GetAppFiles(1, docInfo);{Only Look at the first. Ignore the fact that an additional four might be included.}
with docInfo do
if fType = IndicusFileType then
begin
TheIndex := EmptyIndex;
if ReadInIndex(TheIndex, fname, TheIndex^.OpenVref) then
begin
TheIndex^.OpenFileName := fname;
end;{Redraw the screen.}
ClrAppFiles(thisDoc)
end
else
Error(concat(fName, ' is the wrong file type. '));
end
end; {DoStartup}





procedure SaveAsAnother;
{Miss Mascarade sez that we should go forth and multiply. So why save the old data. Make it new again.}
label
17;
var
fname: str255;
vref: integer;
{The stuff of which files are made...}
NewName: str255;
{Just in case somebody (and I'm not pointing any mice) backs out instead of backing up.}
FileErr: OSErr;
NewVref: integer;
begin
NewName := StandardFile(StandardPut, '', IndicusFileType, NewVref);
if NewName <> '' then
begin
FileErr := Create(NewName, NewVref, IndicusCreatorType, IndicusFileType);
if FileErr <> noErr then
begin
Error(concat('OS FileErr of Type #', Num2Str(FileErr)));
goto 17
end;
if WriteOutIndex(TheIndex, NewName, NewVref) then
with TheIndex^ do
begin
OpenFileName := NewName;
OpenVref := NewVref;

end
end;
17:
end;{SaveAsAnother}

procedure Revert;
{Not much here, just read in the stuff and try again. Oh, yeah, ask the bloke if he really wants to do it.}
var
result: integer;
tempIndex: IndexPtr;
begin
result := 2;
while (result > 1) or (result < 0) do
begin
write('Are you sure you want to revert? (0=yes 1=no)');
readln(result)
end;
if result = 0 then
begin
BlowOutOldIndex(TheIndex);
tempIndex := EmptyIndex;
if ReadInIndex(tempIndex, TheIndex^.OpenFileName, TheIndex^.OpenVref) then
begin
tempIndex^.OpenFileName := TheIndex^.OpenFileName;
tempIndex^.Openvref := TheIndex^.OpenVref;
TheIndex := tempIndex
end; {Error has been signalled already.}
end;
end;{Revert}

procedure NewName;
{Just get a new name for now...}
label
15;
var
fname: str255;
vref: integer;
{The stuff of which files are made...}
result: integer;
{What the dude says...}
FileErr: OSErr;

begin

if (TheIndex <> nil) & TheIndex^.FileChanged then
begin
result := 3;
while (result > 2) or (result < 0) do
begin
write('Save the File? (0=yes, 1=no, 2=cancel operation)');
readln(result)
end;
if result = 0 then
begin
if TheIndex^.OpenFileName = DefaultFileName then
SaveAsAnother{Well, we should get the name...}
else if not (WriteOutIndex(TheIndex, TheIndex^.OpenFileName, TheIndex^.OpenVref)) then
goto 15
end
else if result = 2 then
goto 15
end;

fName := StandardFile(StandardPut, '', IndicusFileType, vref);
if fName = '' then
fName := DefaultFileName
else
begin
FileErr := Create(fName, vref, IndicusCreatorType, IndicusFileType);
if FileErr <> noErr then
begin
Error(concat('OS Error of Type #', Num2Str(FileErr)));
goto 15
end;
end;

TheIndex := CreateNewIndex;
{Add a new one here...}
TheIndex^.OpenFileName := fname;
TheIndex^.Openvref := vref;
15:
end;{NewName}

end.