
{<                             Disk Index Creator                                        >}
{<                               by Peter Wayner                                         >}
{<                              Copyright 1991 Peter Wayner                      >}
{<                                                                                                  >}
{This software may be freely distributed as source code as long as}
{the following conditions are met:}
{1) It is unchanged.}
{2) The copyright notice is included.}
{3) The distribution is not part of a commercial venture. I.E. no}
{money or barter may be exchanged in return for a copy of the software. }
{Also, this product cannot be given away as a "free" bonus for the purchase of}
{another product. It cannot, for instance, be included on a hard disk by a manufacturer}
{in the hopes of enticing more people to buy that disk.}
{4) The complete source code is included with all distributions. No object code-only}
{versions are allowed.}


{ This software may be included in derivative works as long as the}
{following conditions are included:}
{1) All changes to the code are clearly identified.}
{2) This copyright notice is attached.}
{3) The derivative works are not part of any commercial venture.}
{4) The complete source code continues to be distributed with the derivative}
{works. }
{5) The derivative work also contains the same limits on distribution as}
{this code. }

{A short explanation of the copyright notice attached here. }
{Please feel free to study this code, notice its good points, swear about its bad points and use it }
{as a beginning for learning about the subject. You are welcome to create new versions for other }
{machines, add your own bells and whistles and even pass around your improvements. Just don't }
{sell them or give them away free as part of some other commercial deal. If you want to do this,}
{re-write the code from scratch. It could be an improvement. }


{The source code consists of seven units:}
{1) IndexGlobals.p-- These are the global definitions of constants, types and variables. }
{2) MemoryUnit.p-- These are the functions for maintaining the memory structures for}
{ the word and the list heaps. }
{3) FileBreaker.p -- These routines are used to pull apart a file into its words. They call}
{ the routines in MemoryUnit to store them in the index.}
{4) NewDeletion.p-- These routines take a list of files and removes them and all of their}
{ contents from the index.}
{5) FileStructure.p-- These routines maintain the lists of files in the index. }
{6) InAndOut.p-- Some file operation routines for storing the data and retrieving it.}
{7) Central.p-- A cheap, line-oriented user-interface.}

{The source code runs in Symantec's THINK Pascal 3.01. Most, if not all, of the code which is specific to the Macintosh}
{is identified in the comments next to it. To get the code to run on a non-Macintosh would require changing some of the}
{display routines and all of the disk operating commands. Most of the Mac OS routines should be pretty obvious, but if }
{you have any trouble, consult the Inside Macintosh series published by Addison-Wesley.}


{There are a number of compile-time variables used to control various bits of code included for}
{debugging and instrumatation. They are included for your own fun. A brief description is:}
{When GraphHeapGrowth is true, the program opens a window and graphs the}
{memory usage over the words in the index. Good fun. }
{When NameTest is true, the portion of the code which maintains the list of file names will print}
{messages when it changes the structure.}
{When UseProfiler is true, the program will use the THINK Pascal profiler. This will only work}
{on the Macintosh. The software must also include Profiler.lib and Profiler.p in the project file.}
{When TestWordUsage is true, the software watches how many new words are added}
{to the heap with each new file. It is mainly for instrumentation.}
{When TestDeletion is false, the program will print plenty of messages about what}
{it is doing during deletion. For debugging.}

unit IndexGlobals;
interface

const
ListHeapAddressMask = $1FFFFC;{=1/1111/1111/1111/1111/1100}
ListHeapAddressOffset = 10;
ListHeapValueMask = $0FFF;
ListHeapHeadBit = 31;
ListHeapValueAndHeadBitMask = $80000FFF;
{For constructing the list heap.}

LowSixteen = $FFFF;
AlphaHeapAddressMask = $3FFFFF8;
{This is the constant for masking off the lower  twentysevenbits of a longinteger;}
{it is used to pack data in the functions MemoryToNode and NodeToMemory. The goal}
{is to take a 4 byte pointer and reduce it to three bytes. Well, you could just take the}
{lowest-order three bytes and that would be easy, but since the function doesn't use }
{the lowest three bits (all blocks are 8 bytes long) then why not just take the next 24 significant}
{bits.}
{That is, 00000011 11111111 11111111 11111000}

StartingHeapSize = $1000;
{This is the amount of memory a heap initially starts off with.}
HeapIncrement = $1000;
{When a heap runs out of space, it is incremented by this amount.}

ListNodeSize = 4;
AlphaNodeSize = 8;
TwiceListNodeSize = 8;
TwiceAlphaNodeSize = 16;
{The Number of Bytes in a Node and its double.}
NotAlphaEndValue = 'a';
NotEndValue = 255;
{This is what is written in the Node when it is NOT at the end of the unallocated list.}
AlphaEndValue = '#';
EndValue = 0;
{This is what is written in the Node when it IS at the end of the unallocated list.}
EmptyList = 0;
{List pointers are just longintegers which are added to the relative heap. This is the equivalent to nil.}
MaxTreeDepth = 40;
{This is the largest depth the alphabetical trees will drop.}
MaxFileNumber = $ff0;
{Each file on the disk gets a number. This is the largest possible one.}
MinFileNumber = 1;
{Each file on the disk gets a number. This is the smallest possible one.}
FileNodeSize = 4;
{This is the number of bytes in a file node. Used in the file number heap.}
FileBlockSize = 8196;
{This is the number of bytes read in from the file as it is processed. This}
{should be as large as comfortable in the memory for efficency.}
IndicusCreatorType = 'Ind3';
IndicusFileType = 'InIn';
{The first is the type of the indicus application and the second is the file created by it.}
DefaultFileName = 'Untitled';
type
StandardType = (StandardGet, StandardPut);


wordstring = str255;
longptr = ^longint;

shortptr = ^integer;


heaptype = (alpha, list, number, name);

heapptr = ^heap;
heap = record
heaphndl: handle;
{This is a Macintosh specific definition. A handle is a pointer to a pointer of a chunk}
{of memory. This is done because the Mac does its own memory management and may}
{move a chunk. Of course you can still find it because of the first pointer.}
heapsize: longint;
{The number of bytes in the chunk.}
UnallocatedStart: longint;
{This is the pointer to the beginning of the list of unallocated nodes. }
kind: heaptype;
{Currently all but name will use this...}
end;

RootArray = array['A'..'Z'] of longint;
{These are the locations of nodes in the alphaheap which contain the trees associated with each letter.}


{During the update process, it will often be necessary to delete a list of files values from the }
{heap. This is a list that contains them.}
{I will use regular pascal values here because I don't care how well they are packed. Also because}
{they do not need to be relocated and reloaded.}

valuelistptr = ^valuelistnode;
valuelistnode = record
value: integer;
{This is the value.}
DirID: longint;
{This is a value refering to the directory id of the file that will be updated.}
Vref: integer;
{This is the number refering to the mounted volume. It shouldn't get past the scanner.}
next: valuelistptr;
{this is the next...}
end;

filenamestring = string[31];

{This structure is used to store the different files from the directory.}

filenameNodeptr = ^filenamenode;
filenamenode = record
fname: filenamestring;
{This is either the name of a file or a directory.}
directory: boolean;
{True if this node represents a directory, false if it represents a file.}
{This used to be a boolean, but now it is an integer because THINK Pascal}
{seems to really screw things up with too many booleans in the record.}
Closed: boolean;
{This is a directory that is not complete. The program should not add new files to }
{the directory if it encouters them in the later scanning process. It should ignore them.}
{In theory, if this is false (i.e. not closed, or open) then it is false for all decendents.}
next: longint;
{The relative pointer to the next node in the list.}
Above: longint;
{The one which points to this. This could either be a directory pointing down or another file pointing next.}
down: longint;
{If this is a directory, than this is a pointer to another list. Otherwise it is a file number.}
end;

FileNameHeapPtr = ^FileNameHeapType;
FileNameHeapType = record
heaphndl: handle;
{Where the data will go.}
heapsize: longint;
{The size of it all.}
UnAllocatedStart: longint;
{The next record to be allocated.}
TreeRoot: longint;
{Where the rooot is.}
end;
{The path list is used to represent a file and it's path from the root. The first node should be }
{the root of the file system.}

pathlistptr = ^pathlistnode;
pathlistnode = record
fname: filenamestring;
{The name at this level...}
next: pathlistptr;
{The next one in the list...}
end;



IndexPtr = ^index;
index = record
TheNames: FileNameHeapPtr;
{Where the names of the files are stored.}
TheNumbers: HeapPtr;
{Where the mapping table between file numbers and names is held.}
TheWords: HeapPtr;
{Where the words in the index are stored.}
TheLocations: HeapPtr;
{Where the list of files and their numbers are placed.}
TheRoots: RootArray;
{Where the beginnings of the alphabetical heap get placed for fast access.}
LastUpdateDate: longint;
{The Macintosh provides a 32-bit number which is the number }
{of seconds since some day in the past. This is set to the last day to}
{to be created.}
MaxLettersPerWord: integer;
MinLettersPerWord: integer;
{Each index has these values defined so that some one can modify the}
{amount of detail stored away in a very crude way. I would guess that}
{short words count for plenty.}
FileChanged: boolean;
{When it is read in, it is pure, but once an update or an add has been issued, this is flipped.}
OpenFileName: str255;
{This is the file name assigned to this baby.}
OpenVref: integer;
{The volume it comes from...}
OverLapOverKill: boolean;
{This is a setting which controls what happens to a word that is greater than}
{MaxLettersPerWord. If it is true and  MaxLettersPerWord=5, then "SuperDuper" becomes "Super", "uperD",}
{"perDu","erDup" etc. . }
DisposeExtra: boolean;
{Also handles excess verbiage. Prolixity beware. Buckley take cover.}
{Only holds if OverLapOverKill is false. In this case, if true and MaxLettersPerWord}
{is 5, then "SuperDuper" becomes "Super." If false, then both "Super" and "Duper"}
{are entered.}
end;

const
filenamenodesize = sizeof(filenamenode);
{The Names are defined below. They have a different size and this makes it easy to avoid}
{problems when recoding.}
StartingFileNameHeapSize = 200 * filenamenodesize;
{The file name is a separate structure for holding file names.}
{This is essentially 200 times the FileNameNode size.}

FileNameHeapIncrement = 200 * FileNameNodeSize;
{The amount to be increased.}
{This is essentially 200 times the FileNameNode size.}
var
TheIndex: IndexPtr;
{The Central Reason of existance...}
ToBeAdded: valuelistptr;
ToBeKilled: valuelistptr;
{These are two lists which are maintained to determine what should be}
{added and destroyed when the actual index updating is begun.}
WordsAdded: longint;
{A counter that increments everytime a new word is added to the heap.}
{for monitoring purposes.}



{$IFC TestWordUsage}
WordProfile: array[1..MaxTreeDepth] of longint;
{For checking the number of words saved int eh tree.}
NewWordCount: longint;
{For counting calls to FinishTree.}
LocalNewWordCount: longint;
{For counting calls during each file.}
LocalWordCount: longint;
LastWordStart, LastListStart: longint;
{Used to compute node usage.}
{$ENDC}

function Num2Str (num: longint): Str255;
procedure Error (str: str255);


implementation

function Num2Str; {(num: longint): Str255}
var
str: Str255;
begin
NumToString(num, str);
{This is a built-in function in the Macintosh which converts a number to a string.}
Num2Str := str
end;

procedure Error; {(str: str255)}
var
temp: str255;
begin
writeln(str);

writeln('Continue?');
readln(temp);
end;
end.