/*
 *
 *  NETMAP.C    A graphic display of workstations on a NetWare LAN.
 *
 *  Copyright (c) 1990 Barry Nance
 *      All Rights Reserved
 *
 *  Barry Nance
 *  47 Cider Brook Drive
 *  Wethersfield, CT 06109
 *
 *
 *  This software is distributed by Byte Magazine and the BIX
 *  Information Network. You may use this software yourself, and
 *  you may share it with others. You may use the source code as
 *  a learning guide. Please, though: do not try to sell this
 *  software, and do not try to bundle it with other software
 *  that you sell.
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <time.h>
#include <dos.h>
#include <string.h>
#include <bios.h>
#include <graphics.h>
#include <fcntl.h>
#include <io.h>
#include <alloc.h>

#define ESC             27
#define TAB              9
#define BACKTAB         15
#define CR              13
#define TRUE             1
#define FALSE            0
#define ERASE            0
#define INTERVAL         2L
#define HIGHLIGHT        1
#define DE_HIGHLIGHT    -1


/* The OT_XXXX constants are NetWare values for Object Type */

#define OT_USER         0x0100
#define OT_FILESERVER   0x0400

/* Prototypes */

void    InitializeProgram(void);
void    Shutdown(void);
void    GetNodeInfo(void);
void    ShowNodes(void);
void    DrawNode(int HighlightFlag, int ConnectNumber);
int     IPX_Installed(void);
void    LoadFont(void *FontPtr, char *FileName, int FontID);
void    ShowDetail(int ConnectNumber);
void    GetDiagnosticData(int ConnectNumber);
void    GetServerName(int ServerNumber, char *ServerName);
void    GetLocalTarget(unsigned char *dest_network,
                        unsigned char *dest_node,
                        unsigned int   dest_socket,
                        unsigned char *bridge_address);

void    mouse_init(void);
void    show_mouse (void);
void    hide_mouse (void);
void    get_mouse (void);
void    move_mouse (int new_x, int new_y);
int     left_button_down(void);
int     left_button_up(void);
int     right_button_up(void);
void    strfarcpy(char far *d, char *s, int len);
void    strnearcpy(char *d, char far *s, int len);


/* The information we obtain about a node from the  */
/* GetConnectionInformation() NetWare function is   */
/* stored in a NODELIST struct.                     */

struct NODELIST
    {
    unsigned long  ObjectID;
    unsigned int   ObjectType;
    char           ObjectName[9];
    unsigned char  LoginTime[7];
    };

struct DIAG
    {
    int           DiagReceived;
    unsigned long SendPackets;
    unsigned long InboundPackets;
    unsigned long LostPackets;
    unsigned int  PostponedAES;
    unsigned long ListenECBs;
    unsigned int  CannotFindRoute;
    };

struct IPXHEADER
    {
    unsigned int    checksum;
    unsigned int    length;
    unsigned char   transport_control;
    unsigned char   packet_type;
    unsigned char   dest_network_number [4];
    unsigned char   dest_network_node   [6];
    unsigned int    dest_network_socket;
    unsigned char   source_network_number [4];
    unsigned char   source_network_node   [6];
    unsigned int    source_network_socket;
    };

struct SPXHEADER
    {
    unsigned int    checksum;
    unsigned int    length;
    unsigned char   transport_control;
    unsigned char   packet_type;
    unsigned char   dest_network_number [4];
    unsigned char   dest_network_node   [6];
    unsigned int    dest_network_socket;
    unsigned char   source_network_number [4];
    unsigned char   source_network_node   [6];
    unsigned int    source_network_socket;
    unsigned char   connection_control;
    unsigned char   datastream_type;
    unsigned int    source_connection_id;
    unsigned int    dest_connection_id;
    unsigned int    sequence_number;
    unsigned int    acknowledge_number;
    unsigned int    allocation_number;
    };

struct ECB
    {
    void far        *link_address;
    void far        (*event_service_routine)(void);
    unsigned char   in_use;
    unsigned char   completion_code;
    unsigned int    socket_number;
    unsigned int    connection_id;
    unsigned int    rest_of_workspace;
    unsigned char   driver_workspace  [12];
    unsigned char   immediate_address [ 6];
    unsigned int    packet_count;
    struct {
        void far    *address;
        unsigned int length;
        } packet [2];
    };

struct ADDRREQUEST
    {
    unsigned int  Len;
    unsigned char Type;
    unsigned char ConnectionNumber;
    };

struct ADDRREPLY
    {
    unsigned int  Len;
    unsigned char NetworkAddress[12];
    };

struct CONFIGRQST
    {
    struct IPXHEADER IPXHeader;
    unsigned char    ExclAddrCount;
    unsigned char    ExclAddress0[6];
    };

struct CONFIGREPLY
    {
    struct IPXHEADER IPXHeader;
    unsigned char    MajorVersion;
    unsigned char    MinorVersion;
    unsigned int     SPXDiagSocket;
    unsigned char    ComponentCount;
    unsigned char    Components[20];
    };

struct DIAGREQUEST
    {
    struct SPXHEADER SPXHeader;
    unsigned char    ComponentPosition;
    unsigned char    DiagRqstType;
    };

struct DIAGREPLY
    {
    struct SPXHEADER SPXHeader;
    unsigned char    CompletionCode;
    unsigned long    IntervalMarker;
    unsigned long    SendPackets;
    unsigned int     MalformedECBs;
    unsigned long    InboundPackets;
    unsigned long    LostPackets;
    unsigned long    AESevents;
    unsigned int     PostponedAES;
    unsigned int     MaxPossibleSockets;
    unsigned int     MaxOpenSockets;
    unsigned int     OpenSocketFailures;
    unsigned long    ListenECBs;
    unsigned int     CannotCancelECBs;
    unsigned int     CannotFindRoute;
    unsigned int     Reserved[4];
    };

struct CONNECTTABLE
    {
    unsigned char InUseFlag;
    unsigned char OrderNumber;
    unsigned char NetworkNumber[4];
    unsigned char NodeAddress[6];
    unsigned char ServerSocket[2];
    unsigned char RecvTimeout[2];
    unsigned char RoutingNode[6];
    unsigned char PktSeqNum;
    unsigned char ConnectionNumber;
    unsigned char ConnectStatus;
    unsigned char MaxTimeout[2];
    unsigned char Reserved[5];
    };

extern unsigned int _stklen = 0x4000;

/* Global data items */

void far (*IPX)(void);

char far  *ServerNameTable;
char far  *ServerPtr;
struct CONNECTTABLE far  *ConnectTable;
struct CONNECTTABLE far  *ConnectPtr;

struct  NODELIST OldNodes[108];
struct  NODELIST NewNodes[108];
struct  DIAG     Diag;
int     Node_x[108];
int     Node_y[108];

void    *TriplexFontPtr;
void    *SansSerifFontPtr;
void    *SmallFontPtr;
int     Right, Width, Bottom, Height;
int     NodeCount;
int     TabbedItem = -1;
int     DiagError;
unsigned char NetworkAddress[12];

unsigned char mouse_exists   = FALSE;
unsigned char mouse_shared   = FALSE;
int         num_buttons    = 0;
unsigned    mouse_status   = 0;
unsigned    mouse_x        = 0;
unsigned    mouse_y        = 0;

char        PretendServers[48 * 8];
struct      CONNECTTABLE PretendConnections[8];

void    main(int argc, char *argv[])
        {
        int             k;
        long            Then, Now;
        unsigned char   Key, Scan;

        InitializeProgram();
        Then = time(NULL);

        while (1)
            {
            Now = time(NULL);
            if (Now - Then > INTERVAL)
                {
                GetNodeInfo();
                ShowNodes();
                Then = time(NULL);
                }
            if (bioskey(1))
                {
                k    = bioskey(0);
                Key  = k & 0x00FF;
                Scan = (k & 0xFF00) >> 8;
                if (Key == ESC)
                    {
                    Shutdown();
                    }
                if (Key == TAB)
                    {
                    if (TabbedItem > -1 && TabbedItem < 108
                        && Node_x[TabbedItem] != 0)
                            DrawNode(DE_HIGHLIGHT, TabbedItem);
                    do  if (++TabbedItem == 108) TabbedItem = 0;
                        while (Node_x[TabbedItem] == 0);
                    DrawNode(HIGHLIGHT, TabbedItem);
                    }
                else
                if (Key == 0 && Scan == BACKTAB)
                    {
                    if (TabbedItem > -1 && TabbedItem < 108
                        && Node_x[TabbedItem] != 0)
                            DrawNode(DE_HIGHLIGHT, TabbedItem);
                    do  if (--TabbedItem < 0) TabbedItem = 107;
                        while (Node_x[TabbedItem] == 0);
                    DrawNode(HIGHLIGHT, TabbedItem);
                    }
                else
                if (Key == CR)
                    {
                    if (TabbedItem >= 0 && TabbedItem <= 107)
                        ShowDetail(TabbedItem);
                    }
                }
            if (left_button_up())
                {
                if (mouse_y < 25)
                    {
                    Shutdown();
                    }
                for (k=0; k<108; k++)
                    if ( mouse_x > Node_x[k] -  3
                      && mouse_x < Node_x[k] + 90
                      && mouse_y > Node_y[k] -  5
                      && mouse_y < Node_y[k] + 10 )
                        {
                        DrawNode(DE_HIGHLIGHT, TabbedItem);
                        TabbedItem = k;
                        DrawNode(HIGHLIGHT, TabbedItem);
                        ShowDetail(k);
                        break;
                        }
                }
            if (left_button_down())
                {
                for (k=0; k<108; k++)
                    if ( mouse_x > Node_x[k] -  3
                      && mouse_x < Node_x[k] + 90
                      && mouse_y > Node_y[k] -  5
                      && mouse_y < Node_y[k] + 10 )
                        if (TabbedItem != k)
                            {
                            DrawNode(DE_HIGHLIGHT, TabbedItem);
                            DrawNode(HIGHLIGHT, k);
                            TabbedItem = k;
                            break;
                            }
                }
            }
        }

void    InitializeProgram(void)
        {
        int  i, GraphDriver, GraphMode, ErrorCode;
        unsigned int seg, ofs;

        for (i=0; i<108; i++)
            {
            OldNodes[i].ObjectID = 0L;
            OldNodes[i].ObjectType = 0;
            memset(OldNodes[i].ObjectName, 0, 9);
            memset(OldNodes[i].LoginTime,  0, 7);
            NewNodes[i].ObjectID = 0L;
            NewNodes[i].ObjectType = 0;
            memset(NewNodes[i].ObjectName, 0, 9);
            memset(NewNodes[i].LoginTime,  0, 7);
            }

        if (IPX_Installed())
            {
            _AX = 0xEF04;
            geninterrupt(0x21);
            ofs = _SI;
            seg = _ES;
            ServerNameTable = MK_FP(seg, ofs);
            _AX = 0xEF03;
            geninterrupt(0x21);
            ofs = _SI;
            seg = _ES;
            ConnectTable = MK_FP(seg, ofs);
            }
        else
            {
            printf("NetWare not loaded.  We'll just pretend.  (Press a key) ");
            bioskey(0);
            printf("\n");
            randomize();
            ServerNameTable = (void far *) PretendServers;
            ConnectTable    = (void far *) PretendConnections;
            ConnectTable->InUseFlag     = 1;
            ConnectTable->ConnectStatus = 1;
            ConnectTable++;
            ConnectTable->InUseFlag     = 1;
            ConnectTable->ConnectStatus = 1;
            ConnectTable++;
            ConnectTable->InUseFlag     = 1;
            ConnectTable->ConnectStatus = 1;
            strfarcpy(ServerNameTable,        "SERVER1", 8);
            strfarcpy(ServerNameTable+(1*48), "SERVER2", 8);
            strfarcpy(ServerNameTable+(2*48), "SERVER3", 8);
            strfarcpy(ServerNameTable+(3*48), "", 8);
            strfarcpy(ServerNameTable+(4*48), "", 8);
            strfarcpy(ServerNameTable+(5*48), "", 8);
            strfarcpy(ServerNameTable+(6*48), "", 8);
            strfarcpy(ServerNameTable+(7*48), "", 8);
            NewNodes[100].ObjectID = 0L;
            NewNodes[100].ObjectType = OT_FILESERVER;
            strnearcpy(NewNodes[100].ObjectName, ServerNameTable, 8);
            NewNodes[101].ObjectID = 1L;
            NewNodes[101].ObjectType = OT_FILESERVER;
            strnearcpy(NewNodes[101].ObjectName, ServerNameTable+(1*48), 8);
            NewNodes[102].ObjectID = 2L;
            NewNodes[102].ObjectType = OT_FILESERVER;
            strnearcpy(NewNodes[102].ObjectName, ServerNameTable+(1*48), 8);
            for (i=4; i<40; i++)
                {
                NewNodes[i].ObjectID = 1234L;
                NewNodes[i].ObjectType = OT_USER;
                NewNodes[i].LoginTime[0] = 90;
                NewNodes[i].LoginTime[1] = 11;
                NewNodes[i].LoginTime[2] = 12;
                NewNodes[i].LoginTime[3] =  8;
                NewNodes[i].LoginTime[4] = 45;
                NewNodes[i].LoginTime[5] = 15;
                NewNodes[i].LoginTime[6] =  1;
                sprintf(NewNodes[i].ObjectName, "User%4.4d", i+1);
                }
            }

        GraphDriver = DETECT;
        GraphMode   = 0;
        initgraph(&GraphDriver, &GraphMode, "..");
        ErrorCode = graphresult();
        if (ErrorCode != grOk)
            {
            printf("Graphics error: %s\n", grapherrormsg(ErrorCode));
            exit(0);
            };

        LoadFont(TriplexFontPtr,   "trip.chr", TRIPLEX_FONT);
        LoadFont(SansSerifFontPtr, "sans.chr", SANS_SERIF_FONT);
        LoadFont(SmallFontPtr,     "litt.chr", SMALL_FONT);
        settextjustify(LEFT_TEXT, BOTTOM_TEXT);
        settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);
        setbkcolor(BLACK);
        clearviewport();
        mouse_init();

        Right  = Width  = getmaxx();
        Bottom = Height = getmaxy();

        setfillstyle(INTERLEAVE_FILL, BLUE);
        setcolor(BLACK);
        bar(0, 0, Right, Bottom);
        setcolor(WHITE);
        rectangle(0, 0, Right, Bottom);

        setfillstyle(SOLID_FILL, LIGHTGRAY);
        bar(3, 3, Right-3, 24);
        setcolor(YELLOW);
        setlinestyle(SOLID_LINE, 0, THICK_WIDTH);
        rectangle(0, 0, Right, 24);
        setlinestyle(SOLID_LINE, 0, NORM_WIDTH);

        setcolor(BLACK);
        setusercharsize(60, 100, 60, 100);
        settextstyle(TRIPLEX_FONT, HORIZ_DIR, 0);
        outtextxy(10, 18, "Byte/Bix NetWare Map");

        setcolor(BLACK);
        setusercharsize(140, 100, 140, 100);
        settextstyle(SMALL_FONT, HORIZ_DIR, 0);

        if (mouse_exists)
            outtextxy(300, 16,
                "[ESC] to quit.  Click item for detail.");
        else
            outtextxy(250, 16,
                "ESC to quit. Tab (or Shift-Tab), Enter for detail.");

        if (mouse_exists) show_mouse();

        GetNodeInfo();
        ShowNodes();
        }

void    Shutdown(void)
        {
        closegraph();
        printf("NetMap ended.\n\n");
        printf("(c) 1990 Barry Nance.\n");
        printf("Distributed by Byte Magazine and BIX.\n\n");
        exit(0);
        }

void    GetNodeInfo(void)
        {
        union REGS      Regs;
        struct SREGS    SRegs;

        struct
            {
            unsigned int    Len;
            unsigned char   Type;
            unsigned char   ConnectionNumber;
            } RequestBuffer;

        struct
            {
            unsigned int    Len;
            unsigned long   ObjectID;
            unsigned        ObjectType;
            char            ObjectName[48];
            unsigned char   LoginTime[7];
            unsigned int    Reserved;
            } ReplyBuffer;

        int n;

        if (IPX_Installed())
            {
            for (n=0; n<100; n++)
                {
                strcpy(NewNodes[n].ObjectName, "");
                memset(NewNodes[n].LoginTime,  0, 7);
                NewNodes[n].ObjectID   = 0L;
                NewNodes[n].ObjectType = 0;
                Regs.h.ah = 0xe3;
                RequestBuffer.Len = 2;
                RequestBuffer.Type = 0x16;
                RequestBuffer.ConnectionNumber = n + 1;
                ReplyBuffer.Len = sizeof(ReplyBuffer) - 2;
                Regs.x.si = FP_OFF( (void far *) &RequestBuffer);
                SRegs.ds  = FP_SEG( (void far *) &RequestBuffer);
                Regs.x.di = FP_OFF( (void far *) &ReplyBuffer);
                SRegs.es  = FP_SEG( (void far *) &ReplyBuffer);
                int86x(0x21, &Regs, &Regs, &SRegs);
                if (ReplyBuffer.ObjectType == OT_USER)
                    {
                    memcpy(NewNodes[n].ObjectName, ReplyBuffer.ObjectName, 8);
                    memcpy(NewNodes[n].LoginTime,  ReplyBuffer.LoginTime, 7);
                    NewNodes[n].ObjectID   = ReplyBuffer.ObjectID;
                    NewNodes[n].ObjectType = ReplyBuffer.ObjectType;
                    }
                }
            ConnectPtr = ConnectTable;
            for (n=0; n<8; n++)
                {
                strcpy(NewNodes[n+100].ObjectName, "");
                memset(NewNodes[n+100].LoginTime,  0, 7);
                NewNodes[n+100].ObjectID   = 0L;
                NewNodes[n+100].ObjectType = 0;
                if (ConnectPtr->InUseFlag && ConnectPtr->ConnectStatus)
                    {
                    GetServerName(n, NewNodes[n+100].ObjectName);
                    NewNodes[n+100].ObjectType = OT_FILESERVER;
                    NewNodes[n+100].ObjectID   = (long) n + 1;
                    }
                ConnectPtr++;
                }
            return;
            }

        NewNodes[100].ObjectID = 5677L;
        NewNodes[100].ObjectType = OT_FILESERVER;
        strnearcpy(NewNodes[100].ObjectName, ServerNameTable, 8);
        NewNodes[101].ObjectID = 5678L;
        NewNodes[101].ObjectType = OT_FILESERVER;
        strnearcpy(NewNodes[101].ObjectName, ServerNameTable+(1*48), 8);
        NewNodes[102].ObjectID = 5679L;
        NewNodes[102].ObjectType = OT_FILESERVER;
        strnearcpy(NewNodes[102].ObjectName, ServerNameTable+(2*48), 8);

        for (n=4; n<40; n++)
            {
            NewNodes[n].ObjectID = 1234L;
            NewNodes[n].ObjectType = OT_USER;
            NewNodes[n].LoginTime[0] = 90;
            NewNodes[n].LoginTime[1] = 11;
            NewNodes[n].LoginTime[2] = 12;
            NewNodes[n].LoginTime[3] =  8;
            NewNodes[n].LoginTime[4] = 45;
            NewNodes[n].LoginTime[5] = 15;
            NewNodes[n].LoginTime[6] =  1;
            sprintf(NewNodes[n].ObjectName, "User%4.4d", n+1);
            }

        n = random(100);
        if (OldNodes[n].ObjectID != 0L)
            {
            NewNodes[n].ObjectID = 0L;
            NewNodes[n].ObjectType = 0;
            strcpy(NewNodes[n].ObjectName, "");
            }
        else
            {
            NewNodes[n].ObjectID = 1234L;
            NewNodes[n].ObjectType = OT_USER;
            NewNodes[n].LoginTime[0] = 90;
            NewNodes[n].LoginTime[1] = 12;
            NewNodes[n].LoginTime[2] = 13;
            NewNodes[n].LoginTime[3] =  7;
            NewNodes[n].LoginTime[4] = 20;
            NewNodes[n].LoginTime[5] = 25;
            NewNodes[n].LoginTime[6] =  2;
            sprintf(NewNodes[n].ObjectName, "User%4.4d", n+1);
            }
        }

void    ShowNodes(void)
        {
        int i;

        for (i=0; i<108; i++)
            {
            if (NewNodes[i].ObjectID != OldNodes[i].ObjectID)
                {
                if (NewNodes[i].ObjectType == 0)
                    {
                    DrawNode(ERASE, i);
                    NodeCount--;
                    }
                else
                    {
                    DrawNode(DE_HIGHLIGHT, i);
                    NodeCount++;
                    }
                OldNodes[i] = NewNodes[i];
                }
            }
        }

void    DrawNode(int HighlightFlag, int ConnectNumber)
        {
        static int  x, y;
        static int  Ycenter;
        static int  Xseparation, Yseparation;
        static int  Row, Column, Foreground, Background;
        static int  NodeFigure[18];
        static char NodeName[13];

        if (ConnectNumber < 0 || ConnectNumber > 105)
            return;

        if (mouse_exists) hide_mouse();

        Ycenter = Bottom / 2;
        Xseparation = 100;
        if (Bottom < 250) Yseparation = 6;
        else
        if (Bottom < 420) Yseparation = 10;
        else
            Yseparation = 12;

        setwritemode(COPY_PUT);

        switch (HighlightFlag)
            {
            case ERASE:
                Foreground = BLUE;
                Background = BLUE;
                break;
            case DE_HIGHLIGHT:
                Foreground = BLACK;
                Background = LIGHTGRAY;
                break;
            case HIGHLIGHT:
                Foreground = BLACK;
                Background = WHITE;
                break;
            default:
                return;
            }

        if (NewNodes[ConnectNumber].ObjectType == OT_FILESERVER)
            {
            if (HighlightFlag == DE_HIGHLIGHT)
                {
                Foreground = BLACK;
                Background = CYAN;
                }
            else
            if (HighlightFlag == HIGHLIGHT)
                {
                Foreground = BLACK;
                Background = RED;
                }
            Row    = 1;
            Column = (ConnectNumber - 100) % 6;
            if (Column == 0) Column = 2;
            else if (Column == 1) Column = 3;
            else if (Column == 2) Column = 1;
            else if (Column == 3) Column = 4;
            else if (Column == 4) Column = 0;
            else if (Column == 5) Column = 5;
            y = Ycenter;
            x = 10 + (Column * Xseparation);
            }
        else
            {
            Row    = ConnectNumber / 6;
            Column = ConnectNumber % 6;
            if ( (Row & 0x0001) == 0x0000 )
                y = Ycenter + 10 + (Yseparation * 2) + (Yseparation * Row);
            else
                y = Ycenter - (Yseparation * 2) - (Yseparation * Row);
            x = 10 + (Column * Xseparation);
            }

        Node_x[ConnectNumber] = x;
        Node_y[ConnectNumber] = y;

        NodeFigure[0]  = x;    NodeFigure[1]  = y;
        NodeFigure[2]  = x+4;  NodeFigure[3]  = y-4;
        NodeFigure[4]  = x+76; NodeFigure[5]  = y-4;
        NodeFigure[6]  = x+80; NodeFigure[7]  = y;
        NodeFigure[8]  = x+80; NodeFigure[9]  = y+4;
        NodeFigure[10] = x+76; NodeFigure[11] = y+8;
        NodeFigure[12] = x+4;  NodeFigure[13] = y+8;
        NodeFigure[14] = x;    NodeFigure[15] = y+4;
        NodeFigure[16] = x;    NodeFigure[17] = y;

        setcolor(Foreground);
        setlinestyle(SOLID_LINE, 0, NORM_WIDTH);
        setfillstyle(SOLID_FILL, Background);
        fillpoly(9, NodeFigure);

        if (NewNodes[ConnectNumber].ObjectType == OT_FILESERVER)
            sprintf(NodeName, "%3d %s", 
                        ConnectNumber - 99,
                        NewNodes[ConnectNumber].ObjectName);
        else
            sprintf(NodeName, "%3d %s", 
                        ConnectNumber + 1,
                        NewNodes[ConnectNumber].ObjectName);
        setusercharsize(100, 100, 100, 100);
        settextstyle(SMALL_FONT, HORIZ_DIR, 0);
        outtextxy(x+4, y+5, NodeName);

        setcolor(BLACK);
        drawpoly(9, NodeFigure);

        if (mouse_exists) show_mouse();
        }

void    ShowDetail(int ConnectNumber)
        {
        static int  i, x1, y1, x2, y2;
        static int  year, month, day, hour, minute, second, dayofweek;
        static char string[81];
        static char *weekday[] = 
            { "Sunday",     "Monday",   "Tuesday",  "Wednesday",
              "Thursday",   "Friday",   "Saturday" };

        x1 = 50;
        y1 = 50;
        x2 = Right-50;
        y2 = Bottom-50;

        if (mouse_exists) hide_mouse();

        setfillstyle(SOLID_FILL, LIGHTGRAY);
        bar(x1, y1, x2, y2);
        setcolor(BLACK);
        setlinestyle(SOLID_LINE, 0, THICK_WIDTH);
        rectangle(x1, y1, x2, y2);
        setlinestyle(SOLID_LINE, 0, NORM_WIDTH);

        setcolor(BLACK);
        setusercharsize(140, 100, 140, 100);
        settextstyle(SMALL_FONT, HORIZ_DIR, 0);

        if (NewNodes[ConnectNumber].ObjectType == 0)
            {
            sprintf(string, "Connection Number %d not in use.",
                        ConnectNumber+1);
            outtextxy(x1+10, y1+40, string);
            }
        else
            {
            if (NewNodes[ConnectNumber].ObjectType == OT_FILESERVER)
                {
                sprintf(string, "Server name: %s",
                            NewNodes[ConnectNumber].ObjectName);
                outtextxy(x1+10, y1+20, string);
                }
            else
                {
                sprintf(string, "User name: %s",
                            NewNodes[ConnectNumber].ObjectName);
                outtextxy(x1+10, y1+20, string);
                sprintf(string, "Connection Number: %d", ConnectNumber+1);
                outtextxy(x1+10, y1+40, string);
                year      = NewNodes[ConnectNumber].LoginTime[0];
                month     = NewNodes[ConnectNumber].LoginTime[1];
                day       = NewNodes[ConnectNumber].LoginTime[2];
                hour      = NewNodes[ConnectNumber].LoginTime[3];
                minute    = NewNodes[ConnectNumber].LoginTime[4];
                second    = NewNodes[ConnectNumber].LoginTime[5];
                dayofweek = NewNodes[ConnectNumber].LoginTime[6];
                sprintf(string, 
                        "Logged on %s, %d/%2d/%2d, at %2d:%2.2d:%2.2d",
                        weekday[dayofweek], 
                        month, day, year, 
                        hour, minute, second);
                outtextxy(x1+10, y1+60, string);
                }
            outtextxy(x1+10, y1+80, "Getting diagnostic data...");
            GetDiagnosticData(ConnectNumber);
            setfillstyle(SOLID_FILL, LIGHTGRAY);
            bar(x1+9, y1+70, x2-10, y1+85);
            sprintf(string, 
                "%s %2.2X%2.2X%2.2X%2.2X / %2.2X%2.2X%2.2X%2.2X%2.2X%2.2X",
                    "DIAGNOSTIC REPORT FOR NETWORK/NODE:",
                    NetworkAddress[0], NetworkAddress[1],
                    NetworkAddress[2], NetworkAddress[3],
                    NetworkAddress[4], NetworkAddress[5],
                    NetworkAddress[6], NetworkAddress[7],
                    NetworkAddress[8], NetworkAddress[9]);
            outtextxy(x1+10, y1+80, string);
            if (Diag.DiagReceived)
                {
                sprintf(string, " Packets Sent:  %8ld", Diag.SendPackets);
                outtextxy(x1+10, y1+100, string);
                sprintf(string, " Inbound Pkts:  %8ld", Diag.InboundPackets);
                outtextxy(x1+10, y1+120, string);
                sprintf(string, " Lost Packets:  %8ld", Diag.LostPackets);
                outtextxy(x1+10, y1+140, string);
                sprintf(string, "Postponed AES:     %5d", Diag.PostponedAES);
                outtextxy(x1+10, y1+160, string);
                sprintf(string, "  Listen ECBs:  %8ld", Diag.ListenECBs);
                outtextxy(x1+10, y1+180, string);
                sprintf(string, "  Can't Route:     %5d", Diag.CannotFindRoute);
                outtextxy(x1+10, y1+200, string);
                }
            else
                {
                sprintf(string, "Error %d.  Diagnostic data not available.",
                            DiagError);
                outtextxy(x1+20, y1+100, string);
                }
            }

        settextstyle(DEFAULT_FONT, HORIZ_DIR, 0);
        if (mouse_exists)
            outtextxy(x1+75, y2-10,
                "Press a key or click the mouse to continue.");
        else
            outtextxy(x1+75, y2-10,
                "Press a key to continue.");

        if (mouse_exists) show_mouse();

        while (!bioskey(1) && !left_button_up())
            ;
        if (bioskey(1))
            i = bioskey(0);

        if (mouse_exists) hide_mouse();

        setfillstyle(INTERLEAVE_FILL, BLUE);
        setcolor(BLACK);
        bar(5, 30, Right-5, Bottom-5);
        if (mouse_exists) show_mouse();

        for (i=0; i<108; i++)
            {
            OldNodes[i].ObjectID = 0L;
            OldNodes[i].ObjectType = 0;
            memset(OldNodes[i].ObjectName, 0, 9);
            memset(OldNodes[i].LoginTime,  0, 7);
            NewNodes[i].ObjectID = 0L;
            NewNodes[i].ObjectType = 0;
            memset(NewNodes[i].ObjectName, 0, 9);
            memset(NewNodes[i].LoginTime,  0, 7);
            Node_x[i] = 0;
            Node_y[i] = 0;
            }

        GetNodeInfo();
        ShowNodes();
        DrawNode(HIGHLIGHT, TabbedItem);
        }


void    GetDiagnosticData(int ConnectNumber)
        {
        static union  REGS         Regs;
        static struct SREGS        SRegs;
        static struct ADDRREQUEST  AddrRequestBuffer;
        static struct ADDRREPLY    AddrReplyBuffer;
        static struct CONFIGRQST   ConfigRequest;
        static struct CONFIGREPLY  ConfigReply;
        static struct ECB          ListenECB[3];
        static struct ECB          SendECB;
        static struct ECB          EstablishECB;
        static struct SPXHEADER    EstablishHdr;
        static struct DIAGREQUEST  DiagRequest;
        static struct DIAGREPLY    DiagReply;
        static long                RequestTime;
        static int                 i;
        static int                 SPXConnectID;


        DiagError = 0;
        Diag.DiagReceived = FALSE;

/*        IPX/SPX available?                                        */
        if (!IPX_Installed())
            {
            Diag.DiagReceived    = TRUE;
            Diag.SendPackets     = 400L;
            Diag.InboundPackets  = 500L;
            Diag.LostPackets     = 100L;
            Diag.PostponedAES    = 5;
            Diag.ListenECBs      = 600L;
            Diag.CannotFindRoute = 10;
            NetworkAddress[0]  = 0; NetworkAddress[1] = 0;
            NetworkAddress[2]  = 0; NetworkAddress[3] = 1;
            NetworkAddress[4]  =16; NetworkAddress[5] = 0;
            NetworkAddress[6]  = 0; NetworkAddress[7] = 1;
            NetworkAddress[8]  = 2; NetworkAddress[9] = 3;
            NetworkAddress[10] = 0; NetworkAddress[11] = 0;
            return;
            }

        _BX = 0x0010;
        _AX = 0;
        IPX();
        if (_AL == 0)
            {
            DiagError = 1;
            return;
            }


/*        if server, get Internet address from Connect Table        */
        if (NewNodes[ConnectNumber].ObjectType == OT_FILESERVER)
            {
            ConnectPtr = ConnectTable + (ConnectNumber - 100);
            for (i=0; i<12; i++)
                AddrReplyBuffer.NetworkAddress[i]
                    = ConnectPtr->NetworkNumber[i];
            goto UseDiagnosticSocket;
            }


/*        get Internet address for this connection #                */
        Regs.h.ah = 0xe3;
        AddrRequestBuffer.Len = 2;
        AddrRequestBuffer.Type = 0x13;
        AddrRequestBuffer.ConnectionNumber 
                = (unsigned char) ConnectNumber + 1;
        AddrReplyBuffer.Len = sizeof(AddrReplyBuffer) - 2;
        Regs.x.si = FP_OFF( (void far *) &AddrRequestBuffer);
        SRegs.ds  = FP_SEG( (void far *) &AddrRequestBuffer);
        Regs.x.di = FP_OFF( (void far *) &AddrReplyBuffer);
        SRegs.es  = FP_SEG( (void far *) &AddrReplyBuffer);
        int86x(0x21, &Regs, &Regs, &SRegs);


UseDiagnosticSocket:
/*  set up to send to socket 0x0456 (NetWare diag. socket ID)       */
        AddrReplyBuffer.NetworkAddress[10] = 0x04;
        AddrReplyBuffer.NetworkAddress[11] = 0x56;

        for (i=0; i<12; i++)
            NetworkAddress[i] = AddrReplyBuffer.NetworkAddress[i];

/*        open a socket for IPX                                     */
        _DX = 0x4545;
        _BX = 0x0000;
        _AL = 0x00;
        IPX();

/*        open a socket for SPX                                     */
        _DX = 0x4646;
        _BX = 0x0000;
        _AL = 0x00;
        IPX();

/*        post listen ecb; fragment = config reply packet           */
        memset(&ListenECB[0], 0, sizeof(struct ECB));
        ListenECB[0].socket_number = 0x4545;
        ListenECB[0].packet_count = 1;
        ListenECB[0].packet[0].address = (void far *) &ConfigReply;
        ListenECB[0].packet[0].length  = sizeof(struct CONFIGREPLY);
        memset(&ConfigReply, 0, sizeof(struct CONFIGREPLY));
        ConfigReply.IPXHeader.packet_type = 4;
        _ES = FP_SEG( (void far *) &ListenECB[0]);
        _SI = FP_OFF( (void far *) &ListenECB[0]);
        _BX = 0x0004;
        IPX();

/*        send config request packet to target node                 */
        memset(&SendECB, 0, sizeof(struct ECB));
        SendECB.socket_number = 0x4545;
        SendECB.packet_count = 1;
        SendECB.packet[0].address = (void far *) &ConfigRequest;
        SendECB.packet[0].length  = sizeof(struct CONFIGRQST);
        memset(&ConfigRequest, 0, sizeof(struct CONFIGRQST));
        ConfigRequest.IPXHeader.packet_type = 4;
        memcpy(ConfigRequest.IPXHeader.dest_network_number,
                AddrReplyBuffer.NetworkAddress, 12);
        ConfigRequest.ExclAddrCount = 0;
        memset(ConfigRequest.ExclAddress0, 0, 6);
        GetLocalTarget(ConfigRequest.IPXHeader.dest_network_number,
                       ConfigRequest.IPXHeader.dest_network_node,
                       ConfigRequest.IPXHeader.dest_network_socket,
                       SendECB.immediate_address);
        _ES = FP_SEG( (void far *) &SendECB);
        _SI = FP_OFF( (void far *) &SendECB);
        _BX = 0x0003;
        IPX();

/*        wait up to 5 seconds for response (let IPX breathe)       */
/*        if no response: say so, close socket, and exit            */
        RequestTime = time(NULL);
        while (time(NULL) < RequestTime + 5L && ListenECB[0].in_use)
            {
            _BX = 0x000A;
            IPX();
            }

        if (ListenECB[0].in_use || ListenECB[0].completion_code != 0)
            {
            DiagError = 2;
            _BX = 0x0001;
            _DX = 0x4545;
            IPX();
            _BX = 0x0001;
            _DX = 0x4646;
            IPX();
            return;
            }

/*        post 3 SPX listen ecbs; fragment = diag reply packet      */
        for (i=0; i<3; i++)
            {
            memset(&ListenECB[i], 0, sizeof(struct ECB));
            ListenECB[i].socket_number = 0x4646;
            ListenECB[i].packet_count = 1;
            ListenECB[i].packet[0].address = (void far *) &DiagReply;
            ListenECB[i].packet[0].length  = sizeof(struct DIAGREPLY);
            memset(&DiagReply, 0, sizeof(struct DIAGREPLY));
            DiagReply.SPXHeader.packet_type = 5;
            _ES = FP_SEG( (void far *) &ListenECB[i]);
            _SI = FP_OFF( (void far *) &ListenECB[i]);
            _BX = 0x0017;
            IPX();
            }

/*        establish SPX connection with other node                  */
        memset(&EstablishECB, 0, sizeof(struct ECB));
        EstablishECB.socket_number = 0x4646;
        EstablishECB.packet_count = 1;
        EstablishECB.packet[0].address = (void far *) &EstablishHdr;
        EstablishECB.packet[0].length  = sizeof(struct SPXHEADER);
        memset(&EstablishHdr, 0, sizeof(struct SPXHEADER));
        EstablishHdr.packet_type = 5;
        memcpy(&AddrReplyBuffer.NetworkAddress[10],
                    &ConfigReply.SPXDiagSocket, 2);
        memcpy(EstablishHdr.dest_network_number,
                AddrReplyBuffer.NetworkAddress, 12);
        _ES = FP_SEG( (void far *) &EstablishECB);
        _SI = FP_OFF( (void far *) &EstablishECB);
        _BX = 0x0011;
        _AX = 0;
        IPX();

        SPXConnectID = _DX;

        RequestTime = time(NULL);
        while (time(NULL) < RequestTime + 5L && EstablishECB.in_use)
            {
            _BX = 0x000A;
            IPX();
            }

        if (EstablishECB.in_use || EstablishECB.completion_code != 0)
            {
            DiagError = 3;
            _BX = 0x0001;
            _DX = 0x4545;
            IPX();
            _BX = 0x0001;
            _DX = 0x4646;
            IPX();
            return;
            }

/*        send SPX diagnostic request packet to target              */
        memset(&SendECB, 0, sizeof(struct ECB));
        SendECB.socket_number = 0x4646;
        SendECB.packet_count = 1;
        SendECB.packet[0].address = (void far *) &DiagRequest;
        SendECB.packet[0].length  = sizeof(struct DIAGREQUEST);
        memset(&DiagRequest, 0, sizeof(struct DIAGREQUEST));
        DiagRequest.SPXHeader.packet_type = 5;
        DiagRequest.ComponentPosition = 0;
        DiagRequest.DiagRqstType = 1;
        _ES = FP_SEG( (void far *) &SendECB);
        _SI = FP_OFF( (void far *) &SendECB);
        _DX = SPXConnectID;
        _BX = 0x0016;
        IPX();

/*        wait up to 5 seconds for response (let IPX breathe)       */
/*        if no response: say so, close socket, and exit            */
        RequestTime = time(NULL);
        while (time(NULL) < RequestTime + 5L
            && ListenECB[0].in_use
            && ListenECB[1].in_use
            && ListenECB[2].in_use)
                {
                _BX = 0x000A;
                IPX();
                }

        if (!ListenECB[0].in_use) i = 0;
        else 
        if (!ListenECB[1].in_use) i = 1;
        else i = 2;

        if (ListenECB[i].in_use || ListenECB[i].completion_code != 0)
            {
            DiagError = 4;
            _BX = 0x0001;
            _DX = 0x4545;
            IPX();
            _BX = 0x0001;
            _DX = 0x4646;
            IPX();
            return;
            }


/*        fill in DIAG structure                                    */
        Diag.DiagReceived    = TRUE;
        Diag.SendPackets     = DiagReply.SendPackets;
        Diag.InboundPackets  = DiagReply.InboundPackets;
        Diag.LostPackets     = DiagReply.LostPackets;
        Diag.PostponedAES    = DiagReply.PostponedAES;
        Diag.ListenECBs      = DiagReply.ListenECBs;
        Diag.CannotFindRoute = DiagReply.CannotFindRoute;


/*        Terminate SPX connection                                  */
        memset(&EstablishECB, 0, sizeof(struct ECB));
        EstablishECB.socket_number = 0x4646;
        EstablishECB.packet_count = 1;
        EstablishECB.packet[0].address = (void far *) &EstablishHdr;
        EstablishECB.packet[0].length  = sizeof(struct SPXHEADER);
        memset(&EstablishHdr, 0, sizeof(struct SPXHEADER));
        EstablishHdr.packet_type = 5;
        _ES = FP_SEG( (void far *) &EstablishECB);
        _SI = FP_OFF( (void far *) &EstablishECB);
        _DX = SPXConnectID;
        _BX = 0x0013;
        IPX();


/*        close sockets and return                                  */
        _BX = 0x0001;
        _DX = 0x4545;
        IPX();
        _BX = 0x0001;
        _DX = 0x4646;
        IPX();
        }


void    GetLocalTarget(unsigned char *dest_network,
                        unsigned char *dest_node,
                        unsigned int   dest_socket,
                        unsigned char *bridge_address)
        {
        static struct  {
                unsigned char   network_number [4];
                unsigned char   physical_node  [6];
                unsigned int    socket;
                } request_buffer;

        static struct  {
                unsigned char   local_target [6];
                } reply_buffer;

        memcpy(request_buffer.network_number, dest_network, 4);
        memcpy(request_buffer.physical_node, dest_node, 6);
        request_buffer.socket = dest_socket;

        _ES = FP_SEG( (void far *) &request_buffer);
        _SI = FP_OFF( (void far *) &request_buffer);
        _DI = FP_OFF( (void far *) &reply_buffer);
        _BX = 0x0002;
        IPX();
        memcpy(bridge_address, reply_buffer.local_target, 6);
        }

void    GetServerName(int ServerNumber, char *ServerName)
        {
        char Name[10];

        memset(Name, 0, 10);
        ServerPtr  = ServerNameTable;
        ServerPtr += ServerNumber * 48;
        strnearcpy(Name, ServerPtr, 8);
        strcpy(ServerName, Name);
        }

int     IPX_Installed(void)
        {
        union REGS      Regs;
        struct SREGS    SRegs;

        Regs.x.ax = 0x7a00;
        int86x(0x2f, &Regs, &Regs, &SRegs);
        if (Regs.h.al != 0xff) 
            return FALSE;

        IPX = MK_FP(SRegs.es, Regs.x.di);
        return TRUE;
        }

void    LoadFont(void *FontPtr, char *FileName, int FontID)
        {
        int     Handle, FSize;

        Handle = open(FileName, O_RDONLY | O_BINARY);
        if (Handle == -1)
            {
            closegraph();
            printf("Unable to open font file %s.\n", FileName);
            exit(1);
            };

        FSize = (int) filelength(Handle);
        FontPtr = malloc(FSize);
        if (FontPtr == NULL)
            {
            closegraph();
            printf("Unable to allocate memory for font file %s\n", FileName);
            exit(1);
            };

        if (read(Handle, FontPtr, FSize) != FSize)
            {
            closegraph();
            printf("Unable to read font file %s\n", FileName);
            close(Handle);
            exit(1);
            };

        close(Handle);
        if (registerfarbgifont(FontPtr) != FontID)
            {
            closegraph();
            printf("Unable to register font file %s\n", FileName);
            exit(1);
            };
        }

void    mouse_init(void)
        {
        void far * far *iv;

        iv = MK_FP(0x0000, 0x00CC);
        if (FP_SEG(*iv) == 0x0000)
            {
            mouse_exists = 0;
            num_buttons  = 0;
            }
        else
            {
            _AX = 0;
            geninterrupt(0x33);
            mouse_exists = _AX;
            num_buttons  = _BX;
            }
        }

void    show_mouse (void)
        {
        _AX = 1;
        geninterrupt(0x33);
        } 

void    hide_mouse (void)
        {
        _AX = 2;
        geninterrupt(0x33);
        }

void    get_mouse (void)
        {
        unsigned  b, c, d;

        _AX = 3;
        geninterrupt(0x33);
        b = _BX;
        c = _CX;
        d = _DX;
        mouse_status = b;
        mouse_x      = c;
        mouse_y      = d;
        }

void    move_mouse (int new_x, int new_y)
        {
        unsigned  a, c, d;

        a   = 4;
        c   = new_x;
        d   = new_y;
        _AX = a;
        _CX = c;
        _DX = d;
        geninterrupt(0x33);
        }

int     left_button_down(void)
        {
        unsigned    b, c, d;

        _AX = 3;
        geninterrupt(0x33);
        b = _BX; c = _CX; d = _DX;
        mouse_x  = c;
        mouse_y  = d;
        if ( (b & 0x01) == 0x01) return TRUE;
        return FALSE;
        }

int     left_button_up(void)
        {
        unsigned    b, c, d;

        _AX = 6;
        _BX = 0;
        geninterrupt(0x33);
        b = _BX; c = _CX; d = _DX;
        mouse_x = c;
        mouse_y = d;
        if (b > 0) return TRUE;
        return FALSE;
        }

int     right_button_up(void)
        {
        unsigned    b;

        _AX = 6;
        _BX = 1;
        geninterrupt(0x33);
        b = _BX;
        if (b > 0) return TRUE;
        return FALSE;
        }

void    strfarcpy(char far *d, char *s, int len)
        {
        int i;

        for (i=0; i<len; i++)
            {
            *d = *s;
            if (*d == '\0') break;
            d++;
            s++;
            }
        }

void    strnearcpy(char *d, char far *s, int len)
        {
        int i;

        for (i=0; i<len; i++)
            {
            *d = *s;
            if (*d == '\0') break;
            d++;
            s++;
            }
        }

