{<                             Disk Index Creator                                        >}
{<                               by Peter Wayner                                         >}
{<                              Copyright 1991 Peter Wayner                      >}
{<                                                                                                  >}

{This is part of the software. See the copyright notice in IndexGlobals.p}
{for details. This file may not be distributed without this copyright notice}
{or separately from the other files. }
unit Deletion;

{$SETC testdeletion=false}
interface
uses
IndexGlobals, MemoryUnit;



procedure DeleteBadFilesFromHeap (TheIndex: indexptr; BadValues: valuelistptr);
{The heap will contain some bad values. Search and destroy. This function will look for }
{heads of lists. When it finds one it will hand off the work to DeleteBadFilesFromList which}
{will be specially equipt to deal with the sorted nature of these lists...}

{Okay, to delete a node, this method will take a LIST of file values to be deleted. This will ensure}
{that each node in the list heap is only visited TWICE. }
{}
{The process is straightforward. Start at the beginning of the heap and check each node. If it is the }
{head of a list, then scan that list to see if anything should be deleted. Since the list is sorted it is}
{possible to keep a sorted list of 1000 file numbers to be deleted and only visit each node twice-- one}
{time when the node is checked to see if it is a head and the second time when this list it is in is being }
{scanned. }

implementation

var
heapbottom: longint;
{The bottom of the heap.}
CurrentNode: longint;
{Used to scan and look around for a head of a list!}
AlphaBottom: longint;
{This is the bottom of the TheWords which must be accessed occasionally when}
{it is determined that the head of the list must be munged.}
temp: longint;
{a tempvalue for doing stuff...}

{$IFC testdeletion}
CurrentWord: wordstring;
CurrentPosition: integer;
{$ENDC}
procedure DeleteBadFilesFromHeap; {(TheLocations, TheWords: heapptr; BadValues: valuelistptr)}
{The heap will contain some bad values. Search and destroy. This function will look for }
{heads of lists. When it finds one it will hand off the work to DeleteBadFilesFromList which}
{will be specially equipt to deal with the sorted nature of these lists...}
var
i: char;


procedure LocalAlphaPushNode (Node: longint);
{This takes the node at position Node and adds it to the unallocated list.}
{A location version exists to reducing locking and other stuff.}
var
where: longptr;
{The location being removed.}
pr: longint;
val: integer;
{For Pulling off...}
begin
with TheIndex^.TheLocations^ do
begin
where := longptr(alphabottom + Node);
AlphaNodeToMemory(where, Emptylist, UnAllocatedStart, NotAlphaEndValue, false);
UnAllocatedStart := Node;
end
end;{LocalAlphaPushNode}


procedure LocalPushNode (Node: longint);
{This takes the node at position Node and adds it to the unallocated list.}
{A location version exists to reducing locking and other stuff.}
var
where: longptr;
{The location being removed.}
pr: longint;
val: integer;
{For Pulling off...}
begin
with TheIndex^.TheLocations^ do
begin
where := longptr(heapbottom + Node);
ListNodeToMemory(where, UnAllocatedStart, NotEndValue);
UnAllocatedStart := Node;
end
end;{LocalPushNode}


procedure DeleteBadFilesFromList (var FirstStart: longint; BadOnes: valuelistptr);
{Comrades Awake, there are bad members amoung us. Those who have ignored the tenets of }
{are august union. We went purge them! The BadOness list is sorted (and sortid) and must be}
{removed from the list commencing at liststart in TheLocations. Get to it!}
var
HeapValue: longint;
prev: longint;
{Used to get information from the heap.}
ListStart, OldStart: longint;
{The old position which must be saved if it needs to be deleted.}

location: longptr;
temp: longint;
{To save dereferencing.}


procedure PrintNode (location: longint);
var
next: longint;
value: longint;
head: boolean;
{To pull it out of the heap.}

begin
ListMemoryToNode(longptr(heapbottom + location), next, value);
writeln('Node:', location : 5, ' Next:', next : 5, ' Value:', value);

end;{PRintNode}


begin
OldStart := FirstStart;
liststart := OldStart;
ListMemoryToNode(LongPtr(heapbottom + liststart), liststart, HeapValue);
{Note that this assumes that liststart is valid and this function is only called when it is needed!}
while (BadOnes <> nil) & (BadOnes^.value < HeapValue) do
BadOnes := BadOnes^.next;
if (BadOnes <> nil) & (BadOnes^.value = HeapValue) then
begin
repeat
{And the second shall be first. Not the last, but enough for now.}
LocalPushNode(OldStart);
OldStart := liststart;
ListMemoryToNode(LongPtr(heapbottom + liststart), liststart, HeapValue);
FirstStart := OldStart; {Reset the beginning and pass this backwards.}
BadOnes := BadOnes^.next; {Get rid of one of these.}
until (BadOnes^.value <> HeapValue) or (BadOnes = nil) or (OldStart = EmptyList);{"|"= Short circuit or.}

if OldStart <> EmptyList then
begin
location := LongPtr(heapbottom + Oldstart);
temp := location^;
BSET(temp, ListHeapHeadBit);
location^ := temp;
end;
end;

{Okay, at this point, there is either an empty list or we are about to bypass the first value of liststart}
while (OldStart <> EmptyList) & (BadOnes <> nil) do
begin
temp := BadOnes^.value;
if (HeapValue < temp) then
begin
prev := oldStart;
OldStart := liststart;
ListMemoryToNode(LongPtr(heapbottom + liststart), liststart, HeapValue)
end

{Okay, we are either at the end of the list or at a point to delete one of the values from a list.}
else if HeapValue = temp then
{Delete the head of both...}
begin
ChangeSucc(LongPtr(heapbottom + prev), liststart);
{PrintNode(liststart);}
LocalPushNode(oldStart);{Bye,Bye.}
OldStart := liststart;
ListMemoryToNode(LongPtr(heapbottom + liststart), liststart, HeapValue);

BadOnes := BadOnes^.next;
end
else  {(BadOnes^.value < HeapValue) }
BadOnes := BadOnes^.next;

{Our work is done.}
end;


end;{DeleteBadFilesFromList}





{$IFC testdeletion}
procedure DownAndOut (var StartingLocation: longint; LocalPos: integer);

{$ELSEC}
procedure DownAndOut (var StartingLocation: longint);
{$ENDC}
{This procedure recursively checks an Alpha Heap. The idea is that if it is a leaf}
{node, call up the Location Heap deletion routine to clean that list, otherwise,}
{just recurse down.}
var
PreviousLocation, CurrentLocation: longint;
{Two pointers for following along the way.}
CurrentDown, CurrentNext: longint;
CurrentLetter: char;
CurrentLeaf: boolean;
{For looking at the current node.}
GoingDown: longint;
{A pointer for going down the list and receiving the value coming back.}
begin
PreviousLocation := EmptyList;

CurrentLocation := StartingLocation;
while CurrentLocation <> EmptyList do
begin
AlphaMemoryToNode(longptr(CurrentLocation + AlphaBottom), CurrentNext, CurrentDown, CurrentLetter, CurrentLeaf);
if CurrentLeaf then
begin
GoingDown := BSR(CurrentDown, 1);{Adjust for the differences between the sides.}
{$IFC TestDeletion}
currentWord[localpos] := CurrentLetter;
currentWord[localpos] := ' ';
write('At:', CurrentWord : 10);
write('About Delete From:');
PrintList(theIndex^.TheLocations, GoingDown);
writeln;
{$ENDC}
DeleteBadFilesFromList(GoingDown, BadValues);
{$IFC TestDeletion}
write('Coming Back From:');
PrintList(theIndex^.TheLocations, GoingDown);
writeln;
{$ENDC}
end
else
begin
GoingDown := CurrentDown;
{$IFC testdeletion}
currentWord[localpos] := CurrentLetter;
DownAndOut(GoingDown, LocalPos + 1);
currentWord[localpos] := ' ';

{$ELSEC}
DownAndOut(GoingDown);
{$ENDC}
end;
if GoingDown = EmptyList then
{This node needs to be deleted.}
begin
LocalAlphaPushNode(CurrentLocation);
if PreviousLocation = EmptyList then
StartingLocation := CurrentNext
else
AlphaChangeNext(longptr(PreviousLocation + AlphaBottom), CurrentNext);
CurrentLocation := CurrentNext;
end
else
begin
if GoingDown <> BSR(CurrentDown, 1) then
if CurrentLeaf then
AlphaChangeDown(longptr(CurrentLocation + AlphaBottom), BSR(GoingDown, 1))
else
AlphaChangeDown(longptr(CurrentLocation + AlphaBottom), GoingDown);
PreviousLocation := CurrentLocation;
CurrentLocation := CurrentNext;
end;
end

end;{DownAndOut}



begin
with TheIndex^ do
begin
HLock(TheLocations^.heaphndl);
heapbottom := Ord4(TheLocations^.heaphndl^);
HLock(TheWords^.heaphndl);
Alphabottom := Ord4(TheWords^.heaphndl^);

for i := 'A' to 'Z' do
begin
temp := TheRoots[i];
{$IFC testdeletion}
CurrentWord := '                                                   ';
DownAndOut(temp, 1);
{$ELSEC}
DownAndOut(temp);
{$ENDC}

if temp <> TheRoots[i] then
TheRoots[i] := temp;
end;
HUnLock(TheLocations^.Heaphndl);
HUnLock(TheWords^.heaphndl);
end
end;{RecursiveShell}
end.