{<                             Disk Index Creator                                        >}
{<                               by Peter Wayner                                         >}
{<                              Copyright 1991 Peter Wayner                      >}
{<                                                                                                  >}

{This is part of the software. See the copyright notice in IndexGlobals.p}
{for details. This file may not be distributed without this copyright notice}
{or separately from the other files. }


{This is a system that will automagically maintain an index of a hard disk so it is easy to get}
{a pointer to where a particular word might be found in the scheme of things. It should be }
{easy to update and quick to search. There is some tradeoff between these two effects and}
{I'll try to point them out in the comments.}

unit MemoryUnit;

{This is the self-maintaining, garbage collecting heap for creating lists of integers, which in }
{effect, will be file numbers. The structure is deliberately set up to be easy to update. For that}
{reason the lists will be lists of nodes and each node will hold a pointer to its predicessor, a pointer}
{to its successor and the a 16-bit integer. Naturally, the extra pointers will take up some space,}
{in this case 24-bits a piece, but the extra weight will make it easy to add and remove entries }
{to the heap structure. }
{}

{}
{The structure of a node will be eight bytes of the format "PPPSSSVV" where PPP is the 24-bits}
{of the pointer to the predicessor, SSS is the 24-bits of the pointer to the successor and VV}
{is the sixteen bits of integer stored at this node. Note that these will all be relative pointers to}
{the starting pointer of the heap! That means an addition needs to be performed with each access.}
{Assembly coding will speed this dramatically. }

{Macintosh specific notes: This uses a memory conceit known as a "handle." That is a pointer to a }
{pointer of a chunk of memory. The Mac does its own memory management for you so when you }
{want some memory you request a chunk and it gives you a pointer to a pointer of a block the}
{right size. Occasionally it will have to rearrange the chunks and it will do so without asking}
{or telling you that it did it. That is why you have a handle because the first pointer dereference}
{tells you where the chunk actually is. This value is changed whenever it is moved. Two routines,}
{HLock and HUnLock will lock and unlock a handle which is to say, make them off limits to a }
{memory moving routines.}
interface
uses
IndexGlobals;
function AddWord (w1: wordstring; value: longint): boolean;
{This is the core function. It has a major function for finding the location to add}
{the word to, and then it calls the addList routines defined above. It is pretty straightforward.}

procedure ListNodeToMemory (location: longptr; successor: longint; value: longint);
{This procedure packs a pointer, successor, and }
{the value into four bytes and writes them in memory.}
{Note that this is used in all cases except the ones where the head of a list is constructed.}
{That means the highest order bit, number 31, will always be stored as zero.}

{Note the lower three bits of predicessor and successor are assumed to be zero.}
procedure ListMemoryToNode (location: longptr; var successor: longint; var value: longint);
{Reverses NodeToMemory. Unpacks values.}

procedure ChangeSucc (location: longptr; NewSucc: longint);
{This does a simplified NodeToMemory used for the Push Routine below.}
procedure AlphaChangeDown (location: longptr; NewDown: longint);
{This does a simplified NodeToMemory used for the Push Routine below.}
procedure AlphaChangeNext (location: longptr; NewNext: longint);
{This does a simplified NodeToMemory used for in the insertion routines below.}
procedure AlphaNodeToMemory (location: longptr; next, down: longint; letter: char; leaf: boolean);
{This procedure packs the two pointers, next and down, and }
{the value into eight bytes and writes them in memory.}
procedure AlphaMemoryToNode (location: longptr; var next, down: longint; var letter: char; var leaf: boolean);
{Reverses NodeToMemory. Unpacks values.}
function NewHeap (ThisKind: heaptype): heapptr;
{This is something that will return a new, completely initialized heap.}
function InitRoots (AlphaHeap: heapptr): RootArray;
{This baby initializes the roots of the all the trees in the forest.}
{Each tree should have a single branch and a leaf.}
function FindWord (w1: wordstring): longint;
{This function looks in the AlphaHeap for the word, w1, and returns  the}
{ address of the file list attached to it. If it can't find it then it returns }
{the EmptyList.}
procedure PrintList (TheHeap: heapptr; location: longint);
{This prints out a list.}

{$IFC GraphHeapGrowth}
procedure InitGraph;
procedure GraphData;
{This is a procedure for graphing the sizes of the heaps as they grow with added words. }
{It uses the LightSpeed Pascal Draw Window.}
{$ENDC}
implementation

{$IFC GraphHeapGrowth}
const
xoffset = 60;
yoffset = 40;
xlen = 400;
ylen = 200;
stringoffset = 50;
{These are constants which define the size of the Drawing Window.}
mx = 1000000;
my = 1000000;
{These are the maximum x and y values.}
GraphFrequency = 5000;
{How often the graph is updated.}
{$ENDC}
var


MyList1, MyList2, MyList3: longint;
{A testing thing.}
result: boolean;
{Also a testing thing.}
i, rog: integer;
{Used in testing also. }

inputfile: text;
{What is coming in...}

w, st: str255;
{For testing...}

lastx, lasty1, lasty2: longint;
{These are constants to keep track of values between graphs...}

var
{All of these variables are defined globally here because the Think Compiler}
{seems to be having problems with scoping issues!}
LastAddingLocation: longint;
{This will hold the LastLetter Examined by FindLocation. This is used for information passing.}
CurrentNext, CurrentDown: longint;
{The locations of the CurrentNext and down values in the tree.}
CurrentLetter: char;
{Used for unpacking and searching.}
CurrentLeaf: boolean;
{This means we are at a CurrentLeaf node.}
AlphaHeap: HeapPtr;
{A DeReferenced item...}
AlphaBottom: longint;
{This is the bottom of the list.}
NewDown: longint;
{When AddToList is called, it will sometimes change the beginning of the list.}
LocalPosition: integer;
{For moving along the words passed to AddWord.}
listbottom: longint;
{This is the start of the TheHeap which is added to each location and then cast into a pointer.}

var
PreviousLocation: longint;
{This is where the one before last location.}
NewNode: longint;
{Whenever a new node is inserted, this is where it goes.}

prevnext, prevdown: longint;
prevletter: char;
prevleaf: boolean;
{ Used for insertion of a new node}

temp: char;
{To save loop dereferencing.}



{$IFC GraphHeapGrowth}
procedure InitGraph;
{Initializes the LightSpeed Pascal (oops Think Pascal) drawing window for data display}
const
winoffx = 100;
winoffy = 200;
{This governs the opening position.}
texty = 100;
var
r: rect;
{For setting the size of the rectangle.}
begin
SetRect(r, winoffx, winoffy, winoffx + xlen + 2 * xoffset, winoffy + ylen + 2 * yoffset);
SetDrawingRect(r);
SetRect(r, winoffx, texty, winoffx + xlen + 2 * xoffset, winoffy - 3);
SetTextRect(r);
ShowDrawing;
ShowText;
MoveTo(xoffset, yoffset);
LineTo(xoffset, yoffset + ylen);
LineTo(xoffset + xlen, yoffset + ylen);
DrawString(Num2Str(mx));
MoveTo(xoffset - stringoffset, yoffset);
DrawString(Num2Str(my));
lastx := xoffset;
lasty1 := yoffset + ylen;
lasty2 := lasty1;
end;{InitGraph}

procedure GraphData;
{This is a procedure for graphing the sizes of the heaps as they grow with added words. }
{It uses the LightSpeed Pascal Draw Window.}

var
newy1, newy2: longint;
{The new values...}
newx: longint;
begin
newx := xoffset + (WordsAdded * xlen) div mx;
newy1 := yoffset + ylen - (TheIndex^.TheWords^.heapsize * ylen) div my;
newy2 := yoffset + ylen - (TheIndex^.TheLocations^.Heapsize * ylen) div my;
MoveTo(lastx, lasty1);
LineTo(newx, newy1);
MoveTo(lastx, lasty2);
LineTo(newx, newy2);
lastx := newx;
lasty1 := newy1;
lasty2 := newy2;
writeln('Words:', wordsadded : 6, '  AlphaSize:', TheIndex^.TheWords^.heapsize : 7, '  ListSize:', TheIndex^.TheLocations^.heapsize : 7);
end;{GraphData}

{$ENDC}

procedure ListNodeToMemory (location: longptr; successor: longint; value: longint);
{This procedure packs a pointer, successor, and }
{the value into four bytes and writes them in memory.}
{Note that this is used in all cases except the ones where the head of a list is constructed.}
{That means the highest order bit, number 31, will always be stored as zero.}

{Note the lower three bits of predicessor and successor are assumed to be zero.}
var
temp: longint;
{Just used to hold it on the way.}
begin
temp := BSL(BAND(successor, ListHeapAddressMask), ListHeapAddressOffset);
temp := BOR(temp, BAND(value, ListHeapValueMask));
location^ := temp;
end;{ListNodeToMemory}

procedure ListHeadNodeToMemory (location: longptr; successor: longint; value: longint);
{This procedure packs the two pointers, predicessor and successor, and }
{the value into eight bytes and writes them in memory.}

{This is a special version of NodeToMemory which is used whenever a new list is}
{created. It sets a high order bit of the first 4 byte word to be 1 so that the node}
{can be recognized as a head. This will be used later in the deletion routines used }
{to update the heap.}
{That means the highest order bit, number 31, will always be stored as zero.}

{Note the lower three bits of predicessor and successor are assumed to be zero.}
var
temp: longint;
{Just used to hold it on the way.}
begin
temp := BSL(BAND(successor, ListHeapAddressMask), ListHeapAddressOffset);
temp := BOR(temp, BAND(value, ListHeapValueMask));
BSET(temp, ListHeapHeadBit);
location^ := temp;
end;{ListHeadNodeToMemory}



procedure ChangeSucc (location: longptr; NewSucc: longint);
{This does a simplified NodeToMemory used for the Push Routine below.}
var
temp: longint;
{Just used to hold it on the way.}
begin
temp := BAND(ListHeapValueAndHeadBitMask, location^);
temp := BOR(temp, BSL(NewSucc, ListHeapAddressOffset));
location^ := temp;
end;

procedure ListMemoryToNode (location: longptr; var successor: longint; var value: longint);
{Reverses NodeToMemory. Unpacks values.}
var
temp: longint;
begin
temp := location^;
successor := BAND(ListHeapAddressMask, BSR(temp, ListHeapAddressOffset));
value := BAND(temp, ListHeapValueMask);
end;{ListMemoryToNode}

procedure ListHeadMemoryToNode (location: longptr; var successor: longint; var value: longint; var Head: boolean);
{Reverses NodeToMemory. Unpacks values.}
{This is a special value that is used by the deletion routines to check to see if there is a head}
{of the list. This basically means that the predicessor value is a pointer to the AlphaHeap.}
var
temp: longint;
begin
temp := location^;
successor := BAND(ListHeapAddressMask, BSR(temp, ListHeapAddressOffset));
value := BAND(temp, ListHeapValueMask);
head := btst(temp, 31);
end;{ListHeadMemoryToNode}





{------------------------------------------------------------------------}
{------------------------------------------------------------------------}
{------- Now the alphabet tree stuff.}
{------------------------------------------------------------------------}
{------------------------------------------------------------------------}

{This section is where the tree is built for holding the different combinations of letters.}
{It will be a forest of 26 trees, each of which is rooted by one of the letters.}
{The paths from the root to the leaf will spell out a word in the index. Each node}
{will contain two pointers, a boolean indicating whether it is a leaf and a character. }
{In an interior node the first pointer will point to the next node at the same level}
{while the second pointer will point to the next level down. This saves a certain amount}
{of space at the cost of a linear search of each level.}
{In a leaf node, the first pointer will still point to the next node on the same level,}
{but the second one will point to a list beginning in a heap. This list will contain the }
{numbers of files containing this word. }
{Here is a rough picture:}
{ (a)}
{  V}
{  (c) -> (d) -> (t)}
{            V    ------------ a pointer pointing down from (d)}
{            (a) -> (d) -> (o)}
{Finding the word "add" amounts to following the path from the root, "a" down one level}
{and scanning along the list until the second letter, "d" is found. This is recursively done}
{until it is found.}
{The lists are kept in sorted order.}

{First the memory packing routines. These store these pointers in 8 bytes just like the other}
{list elements.}

{Note that occassionally leaf nodes and interior nodes will have the same letter. This is because}
{words are occassionaly prefixes for another. Such as: "it" and "iterate." In this case the}
{leaf node comes AFTER the regular interior node in the horizontal list.}


procedure AlphaNodeToMemory (location: longptr; next, down: longint; letter: char; leaf: boolean);
{This procedure packs the two pointers, next and down, and }
{the value into eight bytes and writes them in memory.}

{Note the lower three bits of next and down are assumed to be zero.}
var
temp: longint;
{Just used to hold it on the way.}
begin

temp := BSL(next, 5);{No need to mask here because it zero extends.}
temp := BOR(temp, BSR(BAND(AlphaHeapAddressMask, down), 19));
location^ := temp;
location := longptr(Ord4(location) + 4);
temp := BOR(ord(letter), BSL(down, 13));
if leaf then
BSET(temp, 9);
location^ := temp;
end;{AlphaNodeToMemory}

procedure AlphaChangeNext (location: longptr; NewNext: longint);
{This does a simplified NodeToMemory used for in the insertion routines below.}
var
temp: longint;
{Just used to hold it on the way.}
begin
temp := BAND(255, location^);
temp := BOR(temp, BSL(NewNext, 5));
location^ := temp;
end;

procedure AlphaChangeDown (location: longptr; NewDown: longint);
{This does a simplified NodeToMemory used for the Push Routine below.}
var
temp: longint;
{Just used to hold it on the way.}
begin
temp := BAND($FFFFFF00, location^);
temp := BOR(temp, BSR(NewDown, 19));
location^ := temp;
location := longptr(Ord4(location) + 4);
temp := BAND($FFFF, location^);
temp := BOR(temp, BSL(NewDown, 13));

location^ := temp;
end;

procedure AlphaStoreListPointer (location: longptr; NewDown: longint);
{On Leaf Nodes, the down section is actually a pointer to a location in }
{the list heap. These pointers are 4 bytes so their offsets can have the }
{third least-significant bit set unlike those which increment by 8. For that}
{reason we will shift them out.}
var
temp: longint;
{Just used to hold it on the way.}
begin
temp := BAND($FFFFFF00, location^);
temp := BOR(temp, BSR(NewDown, 18));
location^ := temp;
location := longptr(Ord4(location) + 4);
temp := BAND($FFFF, location^);
temp := BOR(temp, BSL(NewDown, 14));
location^ := temp;
end;

procedure AlphaReadListPointer (location: longptr; var NewDown: longint);
{On Leaf Nodes, the down section is actually a pointer to a location in }
{the list heap. These pointers are 4 bytes so their offsets can have the }
{third least-significant bit set unlike those which increment by 8. For that}
{reason we will shift them out.}
var
temp: longint;
begin
temp := BSL(BAND($FF, location^), 18);
location := longptr(Ord4(location) + 4);
NewDown := BOR(temp, BSR(BAND($FFFF0000, location^), 14));
end;

procedure AlphaChangeDownAndLeaf (location: longptr; NewDown: longint);
{This does a simplified NodeToMemory used for the Push Routine below.}
var
temp: longint;
{Just used to hold it on the way.}
begin
temp := BAND($FFFFFF00, location^);
temp := BOR(temp, BSR(NewDown, 19));
location^ := temp;
location := longptr(Ord4(location) + 4);
temp := BAND($FFFF, location^);
temp := BOR(temp, BSL(NewDown, 13));
BSET(temp, 9); {And Make Sure it is set to be a leaf!}

location^ := temp;
end;

procedure AlphaMemoryToNode (location: longptr; var next, down: longint; var letter: char; var leaf: boolean);
{Reverses NodeToMemory. Unpacks values.}
var
value: longint;
{For unpacking.}
temp: longint;
begin
temp := location^;
next := BAND(AlphaHeapAddressMask, BSR(temp, 5));
down := BSL(temp, 19);
location := longptr(Ord4(location) + 4);
temp := location^;
down := BAND(AlphaHeapAddressMask, BOR(BSR(temp, 13), down));

if BTST(temp, 9) then
leaf := true
else
leaf := false;

letter := char(BAND($FF, temp));

end;{AlphaMemoryToNode}


function FindWord (w1: wordstring): longint;
{This function looks in the AlphaHeap for the word, w1, and returns  the}
{ address of the file list attached to it. If it can't find it then it returns }
{the EmptyList.}


var
LastFindingLocation: longint;
{This is what is the last location found.}
AlphaBottom: longint;
{This is the bottom of the list.}
next, down: longint;
{The locations of the next and down values in the tree.}
letter: char;
{Used for unpacking and searching.}
leaf: boolean;
{This means we are at a leaf node.}
i: integer;
{Position in the word. I suppose I should make an allusion to Borges.}
temp: char;
{To save loop dereferencing.}
begin
HLock(TheIndex^.TheWords^.heaphndl);
Alphabottom := Ord4(TheIndex^.TheWords^.heaphndl^);
LastFindingLocation := TheIndex^.TheRoots[w1[1]] + AlphaBottom;
AlphaMemoryToNode(LongPtr(LastFindingLocation), next, down, letter, leaf);
i := 2;{At The beginning, more or less.}
{The first thing you do here is check for a leaf because there are single letter words.}


{The Main Loop}
while (i <= MaxTreeDepth) & (w1[i] <> ' ') do {If you want to do interword stuff, replace the space with another character.}
begin
if leaf then {Trouble.}
begin
FindWord := EmptyList;
exit(FindWord);{A Think Pascal extension. Means "Goto the end of the function."}
end;

LastFindingLocation := down;
AlphaMemoryToNode(LongPtr(LastFindingLocation + AlphaBottom), next, down, letter, leaf);
temp := w1[i];
while (next <> EmptyList) & (letter < temp) do
begin
LastFindingLocation := next;
AlphaMemoryToNode(LongPtr(LastFindingLocation + AlphaBottom), next, down, letter, leaf);
end;

if letter = temp then {Trouble.}
i := i + 1
else
begin
FindWord := EmptyList;
exit(FindWord);{A Think Pascal extension. Means "Goto the end of the function."}
end

end;

if i > MaxTreeDepth then
begin
FindWord := down;
Exit(FindWord)
end
else   {We better be at  at a leaf!}
begin
if not (leaf) then
begin
LastFindingLocation := next + AlphaBottom;
AlphaMemoryToNode(LongPtr(LastFindingLocation), LastFindingLocation, down, letter, leaf)
{This just moved us over one.}
end;
if (letter = w1[i - 1]) & leaf then
FindWord := BSR(down, 1)
{For an explanation why, check the data in AlphaStoreListPointer}
{Basically because you need an extra bit of precision with 4 byte nodes.}
else
FindWord := EmptyList;
end;

HUnLock(TheIndex^.TheWords^.heaphndl);

end;{FindWord}

{---------------------------------------------------------------------------}
{---------------------------------------------------------------------------}
{---- Creation and Extension of Heaps.....................................................................}
{---------------------------------------------------------------------------}
{---------------------------------------------------------------------------}


{Warning. If these memory routines are changed, then make sure to change}
{DeleteFromNode which makes some assumptions!}

procedure AlphaFill (MemStart: ptr; BytesToFill: longint; PreviousNode: longint);
{Fill is used to create a list of unallocated pointers. MemStart is the location }
{that Fill starts with and it keeps going for BytesToFill bytes. PreviousNode is}
{the pointer that is allocated from before. This is included this way}
{because this function will be used both at the beginning and when the heap}
{is increased in size.}
{Note BytesToFill must be a multiple of the node size. Initially this is 8 bytes,}
{but it could be changed.}
var
last: ptr;
{This is the last location to be written.}

begin
if BytesToFill > 0 then
begin
last := ptr(Ord4(MemStart) + BytesToFill);
while MemStart <> last do
begin
AlphaNodeToMemory(longptr(MemStart), PreviousNode, (PreviousNode + TwiceAlphaNodeSize), NotAlphaEndValue, false);
PreviousNode := PreviousNode + AlphaNodeSize;
MemStart := ptr(Ord4(MemStart) + AlphaNodeSize);
end;
{Well, we're at the end. Put an EndValue there...}
alphaNodeToMemory(longptr(Ord4(MemStart) - AlphaNodeSize), PreviousNode - AlphaNodeSize, (PreviousNode + AlphaNodeSize), AlphaEndValue, false);

end
end;{AlphaFill}

procedure ListFill (MemStart: ptr; BytesToFill: longint; PreviousNode: longint);
{Fill is used to create a list of unallocated pointers. MemStart is the location }
{that Fill starts with and it keeps going for BytesToFill bytes. PreviousNode is}
{the pointer that is allocated from before. This is included this way}
{because this function will be used both at the beginning and when the heap}
{is increased in size.}
{Note BytesToFill must be a multiple of the node size. Initially this is 8 bytes,}
{but it could be changed.}
var
last: ptr;
{This is the last location to be written.}

begin
if BytesToFill > 0 then
begin
last := ptr(Ord4(MemStart) + BytesToFill);
PreviousNode := PreviousNode + TwiceListNodeSize;
while MemStart <> last do
begin
ListNodeToMemory(longptr(MemStart), PreviousNode, NotEndValue);
PreviousNode := PreviousNode + ListNodeSize;
MemStart := ptr(Ord4(MemStart) + ListNodeSize);
end;
{Well, we're at the end. Put an EndValue there...}

ListNodeToMemory(Longptr(Ord4(MemStart) - ListNodeSize), PreviousNode - ListNodeSize, EndValue);
end
end;{ListFill}

function NewHeap (ThisKind: heaptype): heapptr;
{This is something that will return a new, completely initialized heap.}
{IT will currently work for only alpha and list heaps.}
label
999; {For errors.}
var
tempheap: heapptr;
{To hold it until the end.}
begin
New(tempheap);
with tempheap^ do
begin
heapsize := StartingHeapSize;
heaphndl := NewHandle(StartingHeapSize);
if heaphndl = nil then
begin
Error('Not Enough Memory');
NewHeap := nil;
goto 999;
end;

{Now fill up the memory.}
HLock(heaphndl);
kind := ThisKind;
case ThisKind of
alpha: 
begin
UnAllocatedStart := AlphaNodeSize;
AlphaFill(heaphndl^, heapsize, -Alphanodesize); {The zero here is because it is the start of the list.}
end;
list: 
begin
UnAllocatedStart := ListNodeSize;
ListFill(heaphndl^, heapsize, -Listnodesize); {The zero here is because it is the start of the list.}
end;

end;
HUnlock(heaphndl);
end;
NewHeap := tempHeap;
999:
end;{NewHeap}

function GrowHeap (TheHeap: heapptr): boolean;
{This baby will expand the size of the heap by the HeapIncrement. It is done because }
{The allocation routine has gotten to the end of the line. It must grow the size of }
{the handle, a macintosh specific operation, and then pad it with more stuff.}
{Returns true if successful and false if unsuccessful.}

{This baby is called when a list building routine reaches the end of the line.}
{Therefore it is not necessary to fix the end of the old heap.}
label
999;{ For trouble}
begin
with TheHeap^ do
begin
HUnLock(heaphndl);

SetHandleSize(heaphndl, heapsize + heapincrement);
if MemError <> NoErr then
begin
Error('Out of Memory!');
GrowHeap := false;
goto 999;
end;
HLock(heaphndl);
case kind of
alpha: 
AlphaFill(ptr(Ord4(heaphndl^) + heapsize), heapincrement, heapsize - alphanodesize);
List: 
ListFill(ptr(Ord4(heaphndl^) + heapsize), heapincrement, heapsize - listnodesize);
end;

heapsize := heapsize + heapincrement;
GrowHeap := true;
end;
999:
end;{GrowHeap}

function InitRoots (AlphaHeap: heapptr): RootArray;
{This baby initializes the roots of the all the trees in the forest.}
{Each tree should have a single branch and a leaf.}
var
NewNode, NewNode2: longint;
{Where is the end of this...}
letter: char;
{For storage...}
AlphaBottom: longint;
{The bottom of the heap.}
function LocalPopNode: longint;
{This baby will remove one node from the unallocated list.}
{This version is local to save locking and unlocking as well as make the error signalling easier.}
var
where: longptr;
{The location being removed.}
pr: longint;
val: char;
{For Pulling off...}
leafage: boolean;
begin
with AlphaHeap^ do
begin
where := longptr(Ord4(heaphndl^) + UnAllocatedStart);
LocalPopNode := UnAllocatedStart;


AlphaMemoryToNode(where, pr, UnAllocatedStart, val, leafage);

{This moves the UnAllocatedStart. If val=EndValue (probably 0, check constant declarations)}
{then ran out of space. Do the grow thing.}
if val = AlphaEndValue then
if not (GrowHeap(AlphaHeap)) then
begin
LastAddingLocation := EmptyList;
exit(InitRoots); {The Error has already been signaled by GrowHeap.}
end
else
Alphabottom := Ord4(AlphaHeap^.heaphndl^); {Re-orient this..}
end;

end;{LocalPopNode}
begin
HLock(AlphaHeap^.heaphndl);
Alphabottom := Ord4(AlphaHeap^.heaphndl^);
for letter := 'A' to 'Z' do
begin
NewNode := LocalPopNode;
NewNode2 := LocalPopNode;
if (NewNode = EmptyList) or (NewNode2 = EmptyList) then
begin
Error('Out of Memory!');
HUnLock(AlphaHeap^.heaphndl); {This better not happen here.}
Exit(InitRoots)
end;
AlphaNodeToMemory(LongPtr(AlphaBottom + NewNode), NewNode2, EmptyList, letter, false);
AlphaNodeToMemory(LongPtr(AlphaBottom + NewNode2), EmptyList, EmptyList, letter, true);
InitRoots[letter] := NewNode;
end;
HUnLock(AlphaHeap^.heaphndl);
end;{InitRoots}


function AddToList (TheHeap: heapptr; var ListStart: longint; value: longint): boolean;
{This function adds a value to a list. The lists are kept in sorted order. If ListStart=EmptyList }
{(see the constant declarations) then it creates a new list and returns the value in ListStart.}
label
999;{For trouble.}
var
LookUpLocation: longptr;
{What will be looked up. }
LastLookUpLocation, NewLocation: longint;
{For keeping track because those who do not remember the past are condemned to repeat it.}
pr, su: longint;
val: longint;
{For Pulling off...}
temp: longint;

function LocalPopNode: longint;
{This baby will remove one node from the unallocated list.}
{This version is local to save locking and unlocking as well as make the error signalling easier.}
var
where: longptr;
{The location being removed.}

val: longint;
{For Pulling off...}
temp: longint;
begin
with TheHeap^ do
begin
where := longptr(listbottom + UnAllocatedStart);
LocalPopNode := UnAllocatedStart;

ListMemoryToNode(where, UnAllocatedStart, val);
{This moves the UnAllocatedStart. If val=EndValue (probably 0, check constant declarations)}
{then ran out of space. Do the grow thing.}

if val = 0 then
if not (GrowHeap(TheHeap)) then
begin
AddToList := false;
goto 999; {The Error has already been signaled by GrowHeap.}
end
else
listbottom := Ord4(heaphndl^);
end;

end;{LocalPopNode}


begin
AddToList := true;
HLock(TheHeap^.heaphndl);
listbottom := Ord4(TheHeap^.heaphndl^);
if ListStart = EmptyList then
begin
ListStart := LocalPopNode;
LookUpLocation := longptr(listbottom + ListStart);
ListHeadNodeToMemory(LookUpLocation, EmptyList, value);
{This stores it away immediately and returns the value in ListStart.}


end
else
{Can't stick it at the beginning. I've got to LOOK where to stick this.}
begin
LookUpLocation := longptr(listbottom + ListStart);
ListMemoryToNode(LookUpLocation, su, val);
if val > value then
begin
{Use the special functions to keep the head bits flipped.}
LastLookUpLocation := ListStart; {The Start is FINE.}
ListStart := LocalPopNode;{Get a new head of the list. Note that ListStart is a VAR parameter!}
LookUpLocation := longptr(listbottom + ListStart);
ListHeadNodeToMemory(LookUpLocation, LastLookUpLocation, value); {Now put in the new value.}
LookUpLocation := longptr(listbottom + LastLookupLocation);
temp := LookUpLocation^;
BCLR(temp, ListHeapHeadBit);
LookupLocation^ := temp;

end
else
begin
LastLookUpLocation := ListStart;
while (val < value) and (su <> EmptyList) do
begin
pr := LastLookUpLocation;
LastLookUpLocation := su;
LookUpLocation := longptr(listbottom + su);
ListMemoryToNode(LookUpLocation, su, val);
end;
if (su = EmptyList) and (val < value) then {Add it to the end.}
begin
NewLocation := LocalPopNode;
ChangeSucc(LookUpLocation, NewLocation);{Replace the old pointer.}
LookUpLocation := longptr(listbottom + NewLocation);
ListNodeToMemory(LookUpLocation, EmptyList, value);{Store it away.}

end
else if val > value then
begin
NewLocation := LocalPopNode;

LookUpLocation := longptr(listbottom + NewLocation);
ListNodeToMemory(LookUpLocation, LastLookUpLocation, value);
LookUpLocation := longptr(listbottom + pr);
ChangeSucc(LookUpLocation, NewLocation);

end;
end
end;
HUnLock(TheHeap^.heaphndl);
999:
end;{AddToList}


procedure PrintList (TheHeap: heapptr; location: longint);
{This prints out a list.}
var
listbottom: longint;
{listbottom=start of heap; pr is a sink}
LookUpLocation: longptr;
{A pointer... Just for the sake of it.}
val: longint;
{For printing.}
head: boolean;
begin
HLock(TheHeap^.heaphndl);
listbottom := Ord4(TheHeap^.heaphndl^);
while location <> EmptyList do
begin
LookUpLocation := longptr(listbottom + location);
ListHeadMemoryToNode(LookUpLocation, location, val, head);
if head then
write('H', val : 5, ',')
else
write(val : 5, ',');
end;
HUnLock(TheHeap^.heaphndl);
end;{PrintList}

function AddWord; {(w1: wordstring; value: longint): boolean}
{This is the core function. It has a major function for finding the location to add}
{the word to, and then it calls the addList routines defined above. It is pretty straightforward.}

procedure FindLocation;
{This function looks for a word in the database. If it finds it, then it returns the}
{address of the last letter of the string in the heap. The results are left in }
{LastAddingLocation. If it can't find it, it adds it and then}
{returns the address. This is a full service function that borrows plenty of its}
{capabilities from the FindWord. In fact, if I was doing a good job and programming}
{in an object oriented way, then all of this would be structured a bit better. But Object}
{Compilers aren't necessarily as common... Also, the program needs to keep track of a bit}
{more information here because it is going to be sticking nodes in the middle of lists, not just}
{finding them. This should make FindWord a bit more efficient.}




procedure LocalPushNode (Node: longint);
{This takes the node at position Node and adds it to the unallocated list.}
var
where: longptr;
{The location being removed.}
pr: longint;
val: integer;
{For Pulling off...}
begin
with AlphaHeap^ do
begin
where := longptr(Ord4(heaphndl^) + Node);
AlphaNodeToMemory(where, 0, UnAllocatedStart, NotAlphaEndValue, false);
UnAllocatedStart := Node;
end
end;{LocalPushNode}


function LocalPopNode: longint;
{This baby will remove one node from the unallocated list.}
{This version is local to save locking and unlocking as well as make the error signalling easier.}
var
where: longptr;
{The location being removed.}
pr: longint;
val: char;
{For Pulling off...}
leafage: boolean;
begin
with AlphaHeap^ do
begin
where := longptr(Ord4(heaphndl^) + UnAllocatedStart);
LocalPopNode := UnAllocatedStart;

AlphaMemoryToNode(where, pr, UnAllocatedStart, val, leafage);
{This moves the UnAllocatedStart. If val=EndValue (probably 0, check constant declarations)}
{then ran out of space. Do the grow thing.}
if val = AlphaEndValue then
if not (GrowHeap(AlphaHeap)) then
begin
LastAddingLocation := EmptyList;
exit(FindLocation); {The Error has already been signaled by GrowHeap.}
end
else
Alphabottom := Ord4(AlphaHeap^.heaphndl^); {Re-orient this..}
end;

end;{LocalPopNode}

function FinishWord (w1: wordstring; currentpos: integer; StartingNode: longint): longint;
{This baby will add the rest of the word, when it can't find it in the heap. w1 is the whole}
{word, currentpos is the FIRST letter that can't be found. StartingNode is the node of the last}
{node that could be found. The function returns the last node if it succeeeds. Otherwise it }
{returns a value known as EmptyList. For memory}
{lossage problems.}
var
NewNode: longint;
{What will be added.}
begin
{$IFC TestWordUsage}
LocalNewWordCount := LocalNewWordCount + 1;
NewWordCount := NewWordCount + 1;
{$ENDC}
NewNode := LocalPopNode;
AlphaChangeDown(LongPtr(AlphaBottom + StartingNode), NewNode);

while (currentpos <= MaxTreeDepth) & (w1[currentpos] <> ' ') do
begin
StartingNode := NewNode;
NewNode := LocalPopNode;

if (currentpos = MaxTreeDepth) | (w1[currentpos + 1] = ' ') then
AlphaNodeToMemory(LongPtr(AlphaBottom + StartingNode), EmptyList, NewNode, w1[currentpos], true)
else
AlphaNodeToMemory(LongPtr(AlphaBottom + StartingNode), EmptyList, NewNode, w1[currentpos], false);

{Note that the "|" character here is a short circuit boolean or. So the second test isn't issued if the first is true.}
currentpos := currentpos + 1;
end;

AlphaChangeDownAndLeaf(LongPtr(AlphaBottom + StartingNode), EmptyList);


LocalPushNode(NewNode);
FinishWord := StartingNode

end;{FinishWord}



begin

WordsAdded := WordsAdded + 1;
{Keeps Track of the Number of Calls.}

{$IFC TestWordUsage}
LocalWordCount := LocalWordCount + 1;
{$ENDC}
{$IFC GraphHeapGrowth}
if WordsAdded mod GraphFrequency = 0 then
GraphData;
{$ENDC}
LastAddingLocation := TheIndex^.TheRoots[w1[1]];
AlphaMemoryToNode(LongPtr(LastAddingLocation + AlphaBottom), CurrentNext, CurrentDown, CurrentLetter, CurrentLeaf);
LocalPosition := 2;{At The beginning, more or less.}

if not (CurrentLeaf) & (CurrentDown = EmptyList) & (w1[2] <> ' ') then {A starting position...}
begin
{This only occurs at the beginning when it doesn't find the right stuff.}
LastAddingLocation := FinishWOrd(w1, 2, LastAddingLocation);
exit(FindLocation);
end;


{The Main Loop. At this point, the program has eliminated the possibility that the first CurrentLetter is a CurrentLeaf.}
{The Main loop begins with LocalPosition set to the next letter in the word. The loop will try to find its location in}
{the tree and if it succeeds it will repeat itself. Otherwise, it will insert the location itself and exit}
{completely.}

while (w1[LocalPosition] <> ' ') do {If you want to do interword stuff, replace the space with another character.}
begin
if CurrentLeaf then
{A new node must be inserted between PreviousLocation}
{and LastAddingLocation. Note that the construction of the Root Node makes certain that there is no }
{way that this could fire before PreviousLocation has been set...}
begin
AlphaMemoryToNode(LongPtr(AlphaBottom + PreviousLocation), prevnext, prevdown, prevletter, prevleaf);
NewNode := LocalPopNode;


if prevnext = LastAddingLocation then
AlphaChangeNext(LongPtr(AlphaBottom + PreviousLocation), NewNode)
else
AlphaChangeDown(LongPtr(AlphaBottom + PreviousLocation), NewNode);
AlphaNodeToMemory(LongPtr(AlphaBottom + NewNode), LastAddingLocation, EmptyList, w1[LocalPosition - 1], false);
LastAddingLocation := FinishWord(w1, LocalPosition, NewNode);
Exit(FindLocation)
end;

PreviousLocation := LastAddingLocation;
LastAddingLocation := CurrentDown;

AlphaMemoryToNode(LongPtr(LastAddingLocation + AlphaBottom), CurrentNext, CurrentDown, CurrentLetter, CurrentLeaf);
temp := w1[LocalPosition];
while (CurrentNext <> EmptyList) & (CurrentLetter < temp) do
begin

PreviousLocation := LastAddingLocation;
LastAddingLocation := CurrentNext;
AlphaMemoryToNode(LongPtr(LastAddingLocation + AlphaBottom), CurrentNext, CurrentDown, CurrentLetter, CurrentLeaf);
end;


if CurrentLetter > temp then
begin
NewNode := LocalPopNode;


AlphaMemoryToNode(LongPtr(AlphaBottom + PreviousLocation), prevnext, prevdown, prevletter, prevleaf);
if prevnext = LastAddingLocation then
AlphaChangeNext(LongPtr(AlphaBottom + PreviousLocation), NewNode)
else
AlphaChangeDown(LongPtr(AlphaBottom + PreviousLocation), NewNode);
AlphaNodeToMemory(LongPtr(AlphaBottom + NewNode), LastAddingLocation, EmptyList, temp, false);
LastAddingLocation := FinishWord(w1, LocalPosition + 1, NewNode);
Exit(FindLocation)
end
else if CurrentLetter = temp then
LocalPosition := LocalPosition + 1
else if (CurrentNext = EmptyList) then {Trouble. Note this excludes CurrentLetter=temp.}
begin
NewNode := LocalPopNode;

AlphaChangeNext(LongPtr(LastAddingLocation + AlphaBottom), NewNode);
AlphaNodeToMemory(LongPtr(AlphaBottom + NewNode), EmptyList, EmptyList, temp, false);
LastAddingLocation := FinishWord(w1, LocalPosition + 1, NewNode);
Exit(FindLocation)
end;
{Okay, if we reach here, CurrentLetter should equal w1[LocalPosition] and we are ready to go CurrentDown.}

end;

{That was the end of the MainLoop. When, and if, it reaches this point, it has made it through all}
{MaxTreeDepth locations. This means it is either at the end, or it has reached a place}
{where a CurrentLeaf node should exist, but it doesn't. So Add it.}

if LocalPosition <= MaxTreeDepth then
 {We better be at  at a CurrentLeaf!}
{The first thing you do here is check for a CurrentLeaf because there are single CurrentLetter words . }
if (w1[LocalPosition] = ' ') then {We're at a CurrentLeaf! Or at least we'd better be!}
begin
if not (CurrentLeaf) then
begin
if CurrentNext = EmptyList then
begin {Add in a CurrentLeaf node here.}
NewNode := LocalPopNode;

AlphaNodeToMemory(LongPtr(AlphaBottom + NewNode), CurrentNext, EmptyList, w1[LocalPosition - 1], true);
AlphaChangeNext(LongPtr(AlphaBottom + LastAddingLocation), NewNode);
LastAddingLocation := NewNode;
exit(FindLocation)
end
else
begin
PreviousLocation := LastAddingLocation;
LastAddingLocation := CurrentNext;
AlphaMemoryToNode(LongPtr(LastAddingLocation + AlphaBottom), CurrentNext, CurrentDown, CurrentLetter, CurrentLeaf);
if CurrentLeaf and (CurrentLetter = w1[LocalPosition - 1]) then
{Everything is cool.}
exit(FindLocation)
else
begin
{We must stick a new one in the middle.}
NewNode := LocalPopNode;

AlphaNodeToMemory(LongPtr(AlphaBottom + NewNode), LastAddingLocation, EmptyList, w1[LocalPosition - 1], true);
AlphaChangeNext(LongPtr(AlphaBottom + PreviousLocation), NewNode);
LastAddingLocation := NewNode;
exit(FindLocation)
end
{This just moved us over one. We had better be on a CurrentLeaf node now.}
end
end

end;

end;{FindLocation}


begin
AlphaHeap := TheIndex^.TheWords;
HLock(AlphaHeap^.heaphndl);
Alphabottom := Ord4(AlphaHeap^.heaphndl^);
FindLocation;
if LastAddingLocation = EmptyList then
AddWord := false
else
begin
AlphaReadListPointer(LongPtr(LastAddingLocation + AlphaBottom), CurrentDown);
NewDown := CurrentDown;

if AddToList(TheIndex^.TheLocations, NewDown, value) then
AddWord := True
else
AddWord := False;
if NewDown <> CurrentDown then
AlphaStoreListPointer(LongPtr(LastAddingLocation + AlphaBottom), NewDown);
{This is incase a new list beginning is created.}

end;

HUnLock(AlphaHeap^.heaphndl);

end;{AddWord}



end.